/**********************************************************************/
/* McLib: Library for Many-Core Research Env.    Arch Lab. TOKYO TECH */
/**********************************************************************/
#ifdef SC
#include "sc/MClib.c"
#else

#include <stdlib.h>
#include <string.h>
#include "MClib.h"

static const char *MC_LIB_VER = 
    "Multi-Core library MClib v2.0.0 2013-03-25";

/**********************************************************************/
inline void MC_setidx(int *id, int x)         { setidx(id, x); }
inline void MC_setidy(int *id, int y)         { setidy(id, y); }
inline void MC_setidxy(int *id, int x, int y) { setidxy(id, x, y); }
inline int  MC_getidx(int id)                 { return getidx(id); }
inline int  MC_getidy(int id)                 { return getidy(id); }

/**********************************************************************/
void MC_gettime(unsigned int *t)
{
    *(volatile unsigned int*)t =
        *(volatile unsigned int *)MP_MMAP_TIMER;
}

/**********************************************************************/
int MC_clock(long long unsigned int *clock){
    *clock = *(volatile long long unsigned int *)SC_MMAP_TIMER;
    return 0;
}

/**********************************************************************/
inline void setidy(int* id, int y) {
    *id = (y & 0xff) | (*id & 0xff00);
}

/**********************************************************************/
inline void setidx(int* id, int x) {
    *id = ((x & 0xff) << 8) | (*id & 0xff);
}

/**********************************************************************/
inline void setidxy(int* id, int x, int y) {
    *id = ((x & 0xff) << 8) | (y & 0xff);
}

/**********************************************************************/
inline int getidy(int id) {
    return 0xff & id;
}

/**********************************************************************/
inline int getidx(int id) {
    return (id >> 8) & 0xff;
}

/**********************************************************************/
inline int MC_get_id(int *id) {
    *id = *(int*)SC_MMAP_NODE_ID;
    *id = *id & 0xffff;
    return 1;
}

/**********************************************************************/
inline int MC_get_rank(int *rank) {
    *rank = *(int*)SC_MMAP_NOC_SIZE;
    return 1;
}

/**********************************************************************/
inline int MC_get_physical_id(int *id) {
    *id = *(int*)SC_MMAP_NODE_ID;
    *id = (*id >> 16) & 0xffff;
    return 1;
}


/**********************************************************************/
void new_barrier() {
    static volatile int ready[16];
    int offset;

    int x_dist, y_dist;
    int dstid;
    int id_x, id_y;
    int rank_x, rank_y;

    {
        int id;
        MC_get_id(&id);
        id_x = getidx(id);
        id_y = getidy(id);

        int rank;
        MC_get_rank(&rank);
        rank_x = getidx(rank);
        rank_y = getidy(rank);
    }
    
    void local_sync() {
        MC_dma_put_4b(dstid, (int*)ready + offset, 1);
        while(ready[offset] == 0);
        ready[offset] = 0;
    }

    offset = -1;

    for(x_dist = ((rank_x << 1) - 1) >> 1; x_dist >= 1; x_dist /= 2) {
        offset++;
        int t_x = ((id_x - 1)^x_dist) + 1;
        if(rank_x < t_x) continue;
        setidxy(&dstid, t_x, id_y);
        local_sync();
    }
    
    for(y_dist = ((rank_y << 1) - 1) >> 1; y_dist >= 1; y_dist /= 2) {
        offset++;
        int t_y = ((id_y - 1)^y_dist) + 1;
        if(rank_y < t_y) continue;
        setidxy(&dstid, id_x, t_y);
        local_sync();
    }
}

/**********************************************************************/
void MC_new_barrier() {
    new_barrier();
}

/**********************************************************************/
inline int MC_init(int *id_x, int *id_y, int *rank_x, int *rank_y) {
    int id, rank;
    MC_get_id(&id);
    MC_get_rank(&rank);

    *id_x = getidx(id);
    *id_y = getidy(id);
    *rank_x = getidx(rank);
    *rank_y = getidy(rank);

    if(*id_x == 1 && *id_y==1)
        printf("## %s\n", MC_LIB_VER);

    if (*id_y > *rank_y) {
        switch (*id_x) {
        case CORE_LOCATE_BARRIER_X:
            printf("$$ core(%2d,%2d) : server barrier \n", 
                      *id_x, *id_y);
            MC_server_barrier();
            break;
        case CORE_LOCATE_LOCK_X:
            printf("$$ core(%2d,%2d) : server lock   \n", 
                      *id_x, *id_y);
            MC_server_lock();
            break;
        default:
            while(1);
        }
    }
    else{
        MC_sleep(10000);
    }

    return 0;
}

/**********************************************************************/
void MC_initialize() {
    int id;
    MC_get_id(&id);
    if(id == 0x101)
        printf("## %s\n", MC_LIB_VER);
}

/**********************************************************************/
void MC_finalize() {
    int id;
    MC_get_id(&id);
//    printf("%04x finalize()", id);
}

/**********************************************************************/
void MC_fini() {
    MC_finalize();
}

/**********************************************************************/
int MC_barrier()
{
    volatile int *p = (int *)MEM_BARRIER_WORKER_FLAG;
    static int initialize = 1;
    static int id, rank, remote = 0;
    static int *adr;

    if(initialize){ /***** initialize variables *****/
        initialize = 0;
        MC_get_id(&id);
        MC_get_rank(&rank);
        setidx(&remote, CORE_LOCATE_BARRIER_X);
        setidy(&remote, getidy(rank)+1);
        adr = (int*)MEM_BARRIER_SERVER_BASE +
            (getidx(id)-1) + (getidy(id)-1) * getidx(rank);
    }

    *p = 1;
    MC_dma_put_4b(remote, adr, 1);
    while (*p); // busy wait for the ack
    return 0;
}

/**********************************************************************/
int MC_server_barrier()
{
    int rank;
    int i, x, y, count=0;
    int mx, my, ncore=0;
    volatile int *barrier_flag;

    barrier_flag = (int *)MEM_BARRIER_SERVER_BASE;
    // for(i=0; i<ncore; i++) barrier_flag[i]=0;
    MC_get_rank(&rank);
    mx = getidx(rank);
    my = getidy(rank);
    ncore = mx * my;

    while(1) {
        /***** (1) wait *****/
        do {
            count = 0;
            for(i = 0; i < ncore; i++) {
                if(barrier_flag[i]!=0) count++;
            }
        } while(count!=ncore);

        /***** (2) clear all flag *****/
        for(i = 0; i< ncore; i++) barrier_flag[i]=0;

        /***** (3) send ack *****/
        for(y = 1; y <= my; y++) {
            int remote = 0;
            setidy(&remote, y);
            for(x = 1; x <= mx; x++) {
                setidx(&remote, x);
                MC_dma_put_4b(remote, (int *)MEM_BARRIER_WORKER_FLAG, 0);
            }
        }
    }
    return 0;
}

/**********************************************************************/
int MC_lock(int lock)
{
    volatile int *wait_lock    = (int*)MEM_LOCK_WORKER_FLAG;
    static int initialize = 1;
    static int id, rank, remote = 0;
    static int *adr;

    if(initialize){ /***** initialize variables *****/
        initialize = 0;
        MC_get_rank(&rank);
        MC_get_id(&id);
        adr = (int *)MEM_LOCK_SERVER_BASE +
            (getidx(id)-1) + (getidy(id)-1)*getidx(rank);
        setidxy(&remote, CORE_LOCATE_LOCK_X, getidy(rank) + 1);
    }

    *wait_lock = 1;
    MC_dma_put_4b(remote, adr, 1);
    
    while(*wait_lock);

    return 0;
}

/**********************************************************************/
int MC_unlock(int lock)
{
    int *wait_unlock = (int*)MEM_LOCK_SERVER_FLAG;
    static int initialize = 1;
    static int rank, remote = 0;

    if(initialize){ /***** initialize variables *****/
        initialize = 0;
        MC_get_rank(&rank);
        setidxy(&remote, CORE_LOCATE_LOCK_X, getidy(rank) + 1);
    }

    MC_dma_put_4b(remote, (int*)wait_unlock, 0);
    return 0;
}

/**********************************************************************/
int MC_server_lock() {
    volatile int *lock_request = (int*)MEM_LOCK_SERVER_BASE;
    volatile int *wait_unlock  = (int*)MEM_LOCK_SERVER_FLAG;
    int *wait_lock    = (int*)MEM_LOCK_WORKER_FLAG;
    
    int i=0;
    int x, y;
    int rank, rank_x, rank_y, ncore;
    int remote[128*128]; // should eliminate this array!

    MC_get_rank(&rank);
    rank_x = getidx(rank);
    rank_y = getidy(rank);
    ncore = rank_x * rank_y;

    for(y = 1; y <= rank_y; y++) {
        for(x = 1; x <= rank_x; x++) {
            setidxy(&remote[i], x, y);
            i++;
        }
    }

    i = 0;
    while(1) {
        do { /***** find the lock request *****/
            i++; if(i>=ncore) i=0;
        } while(lock_request[i]==0);
        lock_request[i]=0;

        *wait_unlock = 1;
        MC_dma_put_4b(remote[i], (int*)wait_lock, 0); // give the lock
        while(*wait_unlock);                       // wait the unlock
    }
    return 0;
}

/**********************************************************************/
// There is no type 'bool' on C
// bool MC_dma_put_enable()
int MC_dma_put_enable()
{
    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;
    return (*cmd == DMA_NA);
}
/**********************************************************************/
void MC_dma_put(int remote_id, void *remote, void *local, size_t size,
                int remote_stride, int local_stride) {

    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;

    while(*cmd != DMA_NA);

#ifdef MC_LIB_DEBUG
    if((int)remote & 0x3 || (int)local & 0x3 || size & 0x3 ||
       remote_stride & 0x3 || local_stride & 0x3) {
        printf("#L ERROR lower 2 bits of arguments are not 0 in MC_dma_put()\n");
        exit(1);
    }
#endif

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_id), "i"(SC_MMAP_DMA_PUT_REMOTE_ID));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(size),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_stride),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local_stride),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(DMA_PUT), "i"(SC_MMAP_DMA_PUT_CMD));

    while(*cmd != DMA_NA);
}

/**********************************************************************/
void MC_dma_put_nowait(int remote_id, void *remote, void *local,
                       size_t size, int remote_stride, int local_stride) {

    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;

    while(*cmd != DMA_NA);

#ifdef MC_LIB_DEBUG
    if((int)remote & 0x3 || (int)local & 0x3 || size & 0x3 ||
       remote_stride & 0x3 || local_stride & 0x3) {
        printf("#L ERROR lower 2 bits of arguments are not 0 in MC_dma_put()\n");
        exit(1);
    }
#endif

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_id), "i"(SC_MMAP_DMA_PUT_REMOTE_ID));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(size),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_stride),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local_stride),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(DMA_PUT), "i"(SC_MMAP_DMA_PUT_CMD));

}

/**********************************************************************/
void MC_dma_puts(void *remote, void *local, size_t size,
                 int remote_stride, int local_stride) {
    
    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;

    while(*cmd != DMA_NA);

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(size),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_stride),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local_stride),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(3), "i"(SC_MMAP_DMA_PUT_CMD));

    while(*cmd != DMA_NA);
}

/**********************************************************************/
void MC_dma_put64(int remote_id, uint064_t remote, void *local, size_t size,
                  int local_stride) {
    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;
    uint032_t remote_l = (uint032_t)(remote & 0xffffffff);
    uint032_t remote_h = (uint032_t)(remote >> 32);

    while(*cmd != DMA_NA);

#ifdef MC_LIB_DEBUG
    if((int)remote & 0x3 || (int)local & 0x3 || size & 0x3 ||
       local_stride & 0x3) {
        printf("#L ERROR lower 2 bits of arguments are not 0 in MC_dma_put()\n");
        exit(1);
    }
#endif

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_id), "i"(SC_MMAP_DMA_PUT_REMOTE_ID));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_l),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_h),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(local_stride),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(size),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(DMA_PUT64), "i"(SC_MMAP_DMA_PUT_CMD));

    while(*cmd != DMA_NA);
}

/**********************************************************************/
void MC_dma_put_4b(int remote_id, void *addr, int data) {
    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;
    volatile int _data = data;

    while(*cmd != DMA_NA);

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_id), "i"(SC_MMAP_DMA_PUT_REMOTE_ID));
    __asm__ ("sw %0, %1($0)"
             : :"r"(addr),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(&_data),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(DMA_PUT), "i"(SC_MMAP_DMA_PUT_CMD));

    while(*cmd != DMA_NA);
}

/**********************************************************************/
void MC_dma_put_4b_nowait(int remote_id, void *addr, int data) {
    volatile int *cmd = (int*)SC_MMAP_DMA_PUT_CMD;
    volatile int _data = data;

    while(*cmd != DMA_NA);

    __asm__ ("sw %0, %1($0)"
             : :"r"(remote_id), "i"(SC_MMAP_DMA_PUT_REMOTE_ID));
    __asm__ ("sw %0, %1($0)"
             : :"r"(addr),"i"(SC_MMAP_DMA_PUT_REMOTE_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(&_data),"i"(SC_MMAP_DMA_PUT_LOCAL_ADDR));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_SIZE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_REMOTE_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(4),"i"(SC_MMAP_DMA_PUT_LOCAL_STRIDE));
    __asm__ ("sw %0, %1($0)"
             : :"r"(DMA_PUT), "i"(SC_MMAP_DMA_PUT_CMD));
}

/**********************************************************************/
void _MC_dma_get(int remote_id, volatile void *remote, volatile void *local,
                 size_t size, int remote_stride, int local_stride) {
    static unsigned int send_data[8];
    int my_id;
   
    MC_get_id(&my_id);
  
    send_data[0] = my_id;  
    send_data[1] = (int)remote;  
    send_data[2] = remote_stride;  
    send_data[3] = (int)local;  
    send_data[4] = local_stride;  
    send_data[5] = size;  
    send_data[6] = 0;
    send_data[7] = DMA_PUT;
    
    MC_dma_put(remote_id, (void*)SC_MMAP_DMA_GET_REMOTE_ID, send_data, 8*sizeof(int),
               sizeof(int), sizeof(int));
}

/**********************************************************************/
void MC_dma_get(int remote_id, volatile void *remote, volatile void *local,
                size_t size, int remote_stride, int local_stride) {
    volatile unsigned int *flag = (volatile unsigned int*)SC_MMAP_DMA_FLAG;
    
    do {
        *flag = DMAFLAG_NA;
        _MC_dma_get(remote_id, remote, local, size, 
                    remote_stride, local_stride);
        while(*flag == DMAFLAG_NA);
    } while(*flag != DMAFLAG_ACK);
}

/**********************************************************************/
void _MC_dma_get64(int remote_id, uint064_t remote, volatile void *local,
                   size_t size, int local_stride) {
    static unsigned int send_data[8];
    int my_id;
    
    MC_get_id(&my_id);
    
    send_data[0] = my_id;  
    send_data[1] = (int)local;
    send_data[2] = local_stride;
    send_data[3] = (int)(remote & 0xffffffff);
    send_data[4] = (int)(remote >> 32);
    send_data[5] = size;
    send_data[6] = 0;
    send_data[7] = DMA_PUT;

    MC_dma_put64(remote_id, SC_MMAP_DMA_GET_REMOTE_ID, send_data, 8*sizeof(int),
                 sizeof(int));
}

/**********************************************************************/
void MC_dma_get64(int remote_id, uint064_t remote, volatile void *local, 
                  size_t size, int local_stride) {
    volatile int *dma_flag = (volatile int*)SC_MMAP_DMA_FLAG;
    do {
        *dma_flag = DMAFLAG_NA;
        _MC_dma_get64(remote_id, remote, local, size, local_stride);
        while(*dma_flag == DMAFLAG_NA);
    } while(*dma_flag != DMAFLAG_ACK); // retry if nack recieved
}

/**********************************************************************/
int MC_sleep(long long unsigned int n){
    long long unsigned int t1, t2;
    MC_clock((long long unsigned int*)&t1);
    do {
        MC_clock((long long unsigned int*)&t2);
    } while(t2 - t1 < n);
    return 1;
}

/**********************************************************************/
int MC_infloop(){
    __asm__ ("jr   %0" :: "r"(SC_MMAP_INST_LOOP - 4));
    return 1;
}

/**********************************************************************/
void __barrier(int id, int rank, int depth, int maxdepth,
               int mask, volatile int *flags) {
    int server;

    server = (mask & (id - 0x0101)) + 0x0101;

    if(server == id) {
        int dist = (depth >> 1);
        int child1 = ((id - 0x101) | (dist << 8)) + 0x101; // south
        int child2 = ((id - 0x101) | dist) + 0x101; // west
        int child3 = ((id - 0x101) | dist | (dist << 8)) + 0x101; // south-west

        // recv req
        if(getidx(child1) <= getidx(rank))
            while(flags[1] != depth);
        if(getidy(child2) <= getidy(rank))
            while(flags[2] != depth);
        if(getidx(child3) <= getidx(rank) &&
           getidy(child3) <= getidy(rank))
            while(flags[3] != depth);

        mask = (mask << 1) & 0xfefe;

        // more global
        if(depth < maxdepth)
            __barrier(id, rank, depth << 1, maxdepth, mask, flags);

        {
            void *p = (void*)flags;
            // ack
            if(getidx(child1) <= getidx(rank))
                MC_dma_put_4b(child1, p, ~depth);
            
            if(getidy(child2) <= getidy(rank))
                MC_dma_put_4b(child2, p, ~depth);
            
            if(getidx(child3) <= getidx(rank) &&
               getidy(child3) <= getidy(rank))
                MC_dma_put_4b(child3, p, ~depth);
        }
    }
    else {
        // send req
        int offset = (getidy(server) != getidy(id))*2 +
            (getidx(server) != getidx(id));
        MC_dma_put_4b(server, (int*)flags + offset, depth);
        
        // recv ack
        while(flags[0] != ~depth);
    }
}

/**********************************************************************/
int MC_fast_barrier() {
    int id, rank;
    volatile int flags[4];
    int maxdepth;
    
    MC_get_id(&id);
    MC_get_rank(&rank);
    
    memset((void*)flags, 0, sizeof(int)*4);
    maxdepth = (getidx(rank) > getidy(rank)) ?
        getidx(rank) : getidy(rank);
    
    __barrier(id, rank, 2, getidx(rank), 0xfefe, flags);
    return 1;
}

/**********************************************************************/
#endif
