/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "config.h"
#include <iostream>
#include <iomanip>

/**********************************************************************/
void SystemConfig::clear()
{
    lookahead_rc = fix_vc = fix_priority = credit_fc = skip_lt = false;
    num_channel = 1;
    num_inbuf   = 4;
    max_packet_length = 8;
    cleartable();
}


/**********************************************************************/
void SystemConfig::cleartable()
{
    x_app.clear();
    y_app.clear();
    patable.clear();
    pltable.clear();
    lptable.clear();
    lpslave.clear();
    mstable.clear();
    smtable.clear();
}

/**********************************************************************/
bool SystemConfig::readconfigfile(const char *filename)
{
    cleartable();

    MapParser mp;
    mp.parse(filename);
    if (! mp.valid) {
        mp.puterror();
        return false;
    }
    x_rank  = mp.x_rank;
    y_rank  = mp.y_rank;
    x_app   = mp.x_app;
    y_app   = mp.y_app;

    num_core = num_app = 0;
    for (int i = 0; i < x_rank * y_rank; i++) {
        if (! mp.maps[i].valid)
            continue;
        if (mp.maps[i].x > 0 && mp.maps[i].x <= mp.x_app[mp.maps[i].app] &&
            mp.maps[i].y > 0 && mp.maps[i].y <= mp.y_app[mp.maps[i].app])
            num_core++;
        if (num_app < mp.maps[i].app + 1)
            num_app = mp.maps[i].app + 1;
    }
    lptable.resize(num_app);
    lpslave.resize(num_app);

    for (int i = 0; i < x_rank * y_rank; i++) {
        if (! mp.maps[i].valid)
            continue;
        int  x = (i % x_rank) + 1;
        int  y = (i / x_rank) + 1;
        int  app = mp.maps[i].app;
        uint lid = ID(mp.maps[i].x, mp.maps[i].y);
        uint pid = ID(x, y);
        patable[pid] = app;
        pltable[pid] = lid;
        if (! mp.maps[i].slave)
            lptable[app][lid] = pid;                
        else
            lpslave[app][lid] = pid;
        
        if (lptable[app].find(lid) != lptable[app].end() &&
            lpslave[app].find(lid) != lpslave[app].end()) {
            mstable[lptable[app][lid]] = lpslave[app][lid];
            smtable[lpslave[app][lid]] = lptable[app][lid];
        }
    }
    return true;
}

/**********************************************************************/
void SystemConfig::registerdefaultid(uint pid, uint lid, uint rank,
                                     bool smart, std::string pattern)
{
    if (! smart) {
        patable[pid] = 0;
        pltable[pid] = lid;
        lptable[0][lid] = pid;
    } else {
        uint pid_m, pid_s;
        getsmartcorepids(pid, rank, pattern, pid_m, pid_s);
        patable[pid_m] = 0;
        patable[pid_s] = 0;
        pltable[pid_m] = lid;
        pltable[pid_s] = lid;
        lptable[0][lid] = pid_m;
        lpslave[0][lid] = pid_s;
        mstable[pid_m] = pid_s;
        smtable[pid_s] = pid_m;
    }
}

/**********************************************************************/
void SystemConfig::getsmartcorepids(uint pid, uint rank, std::string &pattern,
                                    uint &pid_m, uint &pid_s)
{
    if        (pattern == "CHECK") {
        if (IDY(pid) % 2 == 1) {
            pid_m = ID(IDX(pid) * 2 - 1, IDY(pid));
            pid_s = ID(IDX(pid) * 2    , IDY(pid));
        } else {
            pid_m = ID(IDX(pid) * 2    , IDY(pid));
            pid_s = ID(IDX(pid) * 2 - 1, IDY(pid));
        }
    } else if (pattern == "BLOCK") {
        pid_m = ID(IDX(pid)            , IDY(pid));
        pid_s = ID(IDX(pid) + IDX(rank), IDY(pid));
    } else  /*(pattern == "INTERLEAVE")*/ {
        pid_m = ID(IDX(pid) * 2 - 1, IDY(pid));
        pid_s = ID(IDX(pid) * 2    , IDY(pid));
    }
}

/**********************************************************************/
bool SystemConfig::setdefaultconfig(int x, int y, bool mem, bool server,
                                    bool smart, std::string pattern)
{
    cleartable();

    x_rank = x;
    y_rank = y;
    if (mem) {
        x_rank++;
        y_rank++;
    }
    if (server)
        y_rank++;
    if (smart)
        x_rank *= 2;

    x_app.push_back(x);
    y_app.push_back(y);
    num_core = x * y;
    num_app  = 1;
    lptable.resize(1);
    lpslave.resize(1);

    int poff = (mem) ? 1 : 0;
    uint rank = ID(x, y);
    for (int i = 1; i <= x; i++)
        for (int j = 1; j <= y; j++)
            registerdefaultid(ID(i + poff, j + poff), ID(i, j),
                              rank, smart, pattern);
    
    if (mem)
        registerdefaultid(ID(1, 1), ID(0, 0), rank, smart, pattern);
    if (server) {
        registerdefaultid(ID(1 + poff, y + 1 + poff), ID(1, y + 1),
                          rank, smart, pattern);
        registerdefaultid(ID(2 + poff, y + 1 + poff), ID(2, y + 1),
                          rank, smart, pattern);
    }
    return true;
}

/**********************************************************************/
void SystemConfig::print()
{
    std::cout << "## Physical Size     : "
              << x_rank << 'x' << y_rank << std::endl;
    for (int i = 0; i < num_app; i++)
        std::cout << "## Logical Size App " << i << ": " 
                  << x_app[i] << 'x' << y_app[i] << std::endl;

    for (int y = 1; y <= y_rank; y++) {
        std::cout << "## ";
        for (int x = 1; x <= x_rank; x++) {
            uint pid = ID(x, y);
            if (isvalid(pid)) {
                std::cout << "[" << (char) ('a' + getaid(pid))
                          << std::setw(2) << IDX(getlid(pid)) << ','
                          << std::setw(2) << IDY(getlid(pid))
                          << ((ismaster(pid)) ? 'm' : 
                              (isslave (pid)) ? 's' : ' ') << ']';
            } else {
                std::cout << "[-------]";
            }
        }
        std::cout << std::endl;
    }
}

/**********************************************************************/
