/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "mainmemory.h"

/* MainMemory                                                         */
/**********************************************************************/
MainMemory::MainMemory()
{
    pagetable = NULL;
    lpagetable = NULL;
}

/**********************************************************************/
void MainMemory::init(ullint mem_size)
{
    clear();
    this->mem_size = (mem_size) ? mem_size : MEM_SIZE_DEF;
    use_lpage = (this->mem_size > 0x80000000);
    page_size = PAGE_SIZE_DEF;
    lpage_size = LPAGE_SIZE_DEF;
    initpagetable();
}

/**********************************************************************/
void MainMemory::clear()
{
    if (pagetable) {
        for (ullint i = 0; i < npage; i++)
            if (pagetable[i])
                delete [] pagetable[i];
        delete [] pagetable;
    }
    if (lpagetable) {
        for (ullint i = 0; i < nlpage; i++)
            if (lpagetable[i]) {
                for (ullint j = 0; j < npage; j++)
                    if (lpagetable[i][j])
                        delete [] lpagetable[i][j];
                delete [] lpagetable[i];
            }
        delete [] lpagetable;
    }
}

/**********************************************************************/
void MainMemory::initpagetable()
{
    if (use_lpage) {
        nlpage = (mem_size + lpage_size - 1) / lpage_size;
        npage  = lpage_size / page_size;
        mem_size = lpage_size * nlpage;
        
        lpagetable = new uint032_t**[nlpage];
        for (ullint i = 0; i < nlpage; i++) 
            lpagetable[i] = NULL;
    } else {
        nlpage = 1;
        npage = (mem_size + page_size - 1) / page_size;
        mem_size = page_size * npage;
        pagetable = new uint032_t*[npage];
        for (ullint i = 0; i < npage; i++) 
            pagetable[i] = NULL;
    }
}

/**********************************************************************/
inline uint032_t** MainMemory::getlpage(uint064_t addr)
{
    if (!use_lpage)
        return NULL;
    if (lpagetable[addr / lpage_size])
        return lpagetable[addr / lpage_size];

    uint032_t **lpage = new uint032_t*[npage];
    for (uint j = 0; j < npage; j++)
        lpage[j] = NULL;
    lpagetable[addr / lpage_size] = lpage;
    return lpage;
}

/**********************************************************************/
inline uint032_t* MainMemory::getpage(uint064_t addr)
{
    uint032_t **lpage = (use_lpage) ? getlpage(addr) : pagetable;
    if (lpage[(addr / page_size) % npage])
        return lpage[(addr / page_size) % npage];

    uint032_t *page = new uint032_t[page_size / sizeof(uint032_t)];
    for (uint i = 0; i < page_size / sizeof(uint032_t); i++)
        page[i] = 0;
    lpage[(addr / page_size) % npage] = page;
    return page;
}

/************************************************************************/
inline uint032_t* MainMemory::getrealaddr(uint064_t addr)
{
    uint032_t *page = getpage(addr);
    page += (addr % page_size) / sizeof(uint032_t);
    return page;
}

/**********************************************************************/
int MainMemory::read(uint064_t addr, uint size, void *data)
{
    if (addr >= mem_size)
        return -1;
    uint032_t *mem = getrealaddr(addr);
    if (size == 1) {
        int offset = (addr & 0x3) * 8;
        *(uint008_t *) data = (*mem >> offset) & 0xff;
    } else if (size == 2) {
        int offset = (addr & 0x2) * 8;
        *(uint016_t *) data = (*mem >> offset) & 0xffff;
    } else if (size == 4) {
        *(uint032_t *) data = *mem;
    } else if (size == 8) {
        *(uint064_t *) data = (((uint064_t) mem[1]) << 32) | mem[0];
    } else {
        return -1;
    }
    return 0;
}

/**********************************************************************/
int MainMemory::write(uint064_t addr, uint size, uint064_t data)
{
    if (addr >= mem_size)
        return -1;
    uint032_t *mem = getrealaddr(addr);
    if (size == 1) {
        int offset = (addr & 0x3) * 8;
        uint032_t ins  = (data & 0xff) << offset;
        uint032_t mask = 0xff << offset;
        *mem = ins | (*mem & ~mask);
    } else if (size == 2) {
        int offset = (addr & 0x2) * 8;
        uint032_t ins  = (data & 0xffff) << offset;
        uint032_t mask = 0xffff << offset;
        *mem = ins | (*mem & ~mask);
    } else if (size == 4) {
        *mem = data;
    } else if (size == 8) {
        mem[0] = (uint032_t) data;
        mem[1] = (uint032_t) (data >> 32);
    } else {
        return -1;
    }
    return 0;
}

/************************************************************************/
void MainMemory::print()
{
    uint032_t *page;
    uint032_t **lpage;
    uint064_t over32 = 0x100000000ull / lpage_size;
    bool allzero[page_size / 16 + 2];
    printf("Memory size: 0x%llx\n", mem_size);
    for (ullint i = 0; i < nlpage; i++) {
        if (use_lpage) {
            if ((lpage = lpagetable[i]) == NULL)
                continue;
        } else {
            lpage = pagetable;
        }
        for (ullint j = 0; j < npage; j++) {
            if ((page = lpage[j]) == NULL)
                continue;
            printf("[MEMORY BLOCK: 0x%08llx]", i * npage + j);
            
            allzero[0] = allzero[page_size / 16 + 1] = false;
            for (ullint k = 0; k < page_size / 16; k++) {
                ullint l = k * 4;
                allzero[k + 1] = ((page[l  ] | page[l+1] |
                                   page[l+2] | page[l+3]) == 0);
            }
            for (ullint k = 0; k < page_size / 16; k++) {
                if (allzero[k + 1]) {
                    if (!allzero[k] && allzero[k + 2])
                        printf("\n...");
                    if ( allzero[k] || allzero[k + 2])
                        continue;
                }
                putchar('\n');
                if (i >= over32)
                    printf("%llx ", (i * lpage_size) >> 32);
                printf("%08x: ", (uint)
                       (i * lpage_size + j * page_size + k * 16));
                for (ullint l = k * 4; l < k * 4 + 4; l++)
                    printf("%02x%02x%02x%02x ",
                           page[l] & 0xff , (page[l] >> 8) & 0xff,
                           (page[l] >> 16) & 0xff , (page[l] >> 24) & 0xff);
            }
            printf("\n\n");
        }
    }
}
/************************************************************************/
