/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "mapparser.h"

/**********************************************************************/
void MapParser::clear()
{
    valid = false;
    x_rank = y_rank = 0;
    x_app.clear();
    x_app.resize(26, 0);
    y_app.clear();
    y_app.resize(26, 0);
    maps.clear();    
    errstr.str("");
    errstr.clear();
}

/**********************************************************************/
void MapParser::parse(const char *filename)
{
    clear();

    std::ifstream ifs(filename);
    std::string line;
    if (ifs.fail()) {
        errstr << "!! failed to open file: " << filename << std::endl;
        return;
    }
    bool extend_mode = false;
    std::getline(ifs, line);
    while (! ifs.eof()) {
        std::istringstream iss(line);
        if (line.find("SIMMC_EXTENDED_MAPPING_FILE") != std::string::npos) {
            extend_mode = true;
        } else if (line[0] == '[') {
            std::vector<PLMap> map_line = parseline(iss, extend_mode);
            maps.insert(maps.end(), map_line.begin(), map_line.end());
        } else {
            parsesize(iss, extend_mode);
        }
        std::getline(ifs, line);
    }

    if (sanitycheck())
        valid = true;
}

/**********************************************************************/
std::vector<PLMap> MapParser::parseline(std::istringstream &iss, 
                                         bool extend_mode)
{
    std::vector<PLMap> result;
    char c0, ca, cb, cx, cy;
    int ix, iy;
    iss >> c0;
    while (!iss.eof() && c0 == '[') {
        iss.get(ca);
        if (ca != '-') {
            if (extend_mode) {
                (iss >> ix >> c0 >> iy).get(cb);
            } else {
                (iss >> cx >> c0 >> cy).get(cb);
                ix = (int) cx - '0';
                iy = (int) cy - '0';
            }
            ca = (ca == ' ') ? 'a' : ca;
            cb = (cb == 's') ? 's' : 'm';
            if (ix >= 0 && iy >= 0 && ca >= 'a' && ca <= 'z')
                result.push_back(PLMap(ix, iy, ca - 'a', cb == 's'));
        } else {
            result.push_back(PLMap());
        }
        while (!iss.eof() && ca != ']')
            iss >> ca;
        iss >> c0;
    }
    return result;
}

/**********************************************************************/
void MapParser::parsesize(std::istringstream &iss, bool extend_mode)
{
    char c0, c1, cx, cy;
    int ix, iy;
    if (extend_mode) {
        iss >> ix >> c0 >> iy >> c1;
    } else {
        iss >> cx >> c0 >> cy >> c1;
        ix = (int) cx - '0';
        iy = (int) cy - '0';
    }
    if (ix > 0 && iy > 0 && c0 == 'x') {
        if (c1 == 'P') {
            x_rank = ix;
            y_rank = iy;
            iss >> c0;
        } else if (c1 >= 'a' && c1 <= 'z') {
            x_app[c1 - 'a'] = ix;
            y_app[c1 - 'a'] = iy;
        }
    }
}

/**********************************************************************/
bool MapParser::sanitycheck()
{
    bool sane = true;
    int nodes = x_rank * y_rank;
    maps.resize(nodes);
    // size
    if (nodes == 0) {
        errstr << "!! Physical Size is invalid." << std::endl;
        return false;
    }
    // duplicate
    for (int i = 0; i < nodes; i++) {
        if (! maps[i].valid)
            continue;
        for (int j = i + 1; j < nodes; j++) {
            if (maps[i] == maps[j]) {
                errstr << "!! Logical Node ";
                maps[i].print(errstr);
                errstr << "is duplicated" << std::endl;
                sane = false;
                break;
            }
        }
    }
    // missing
    for (int a = 0; a < 26; a++) {
        if (x_app[a] * y_app[a] == 0)
            continue;
        for (int x = 1; x <= x_app[a]; x++) {
            for (int y = 1; y <= y_app[a]; y++) {
                PLMap comp(x, y, a);
                bool hit = false;
                for (int i = 0; i < nodes; i++)
                    if (maps[i] == comp)
                        hit = true;
                if (! hit) {
                    errstr << "!! Logical Node ";
                    comp.print(errstr);
                    errstr << "is missing" << std::endl;
                    sane = false;
                }
            }
        }
    }
    return sane;
}

/**********************************************************************/
void MapParser::print()
{
    std::cout << "Physical W:" << x_rank << ", H:" << y_rank << std::endl;
    for (int i = 0; i < 26; i++) {
        if (x_app[i] * y_app[i] == 0)
            continue;
        std::cout << "Logical App:" << i << ", W:" << x_app[i] 
                  << ", H:" << y_app[i] << std::endl;
    }
    for (int y = 0; y < y_rank; y++) {
        for (int x = 0; x < x_rank; x++)
            maps[y * x_rank + x].print();
        std::cout << std::endl;
    }
}

/**********************************************************************/
