/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS                         *
 * Copyright (C) 2008-2012 Arch Lab. Tokyo Institute of Technology    *
 *  Dual licensed under New BSD or GPL v2 licenses.                   *
 *  See COPYING file on the base directory for more information.      *
 **********************************************************************/
#include "syscallplus.h"

#ifdef BASE_MODE_FLOAT
#include "../core/float.h"
#else
#include "../core/define.h"
#endif

#define ROUND_UP(a) ((a + (MM_PAGE_SIZE - 1)) & ~(MM_PAGE_SIZE - 1))

/* SystemCallEmulator *************************************************/
/**********************************************************************/
void SystemCallEmulator::init (MemoryController *mc, uint032_t hp, 
                               uint032_t sp)
{
    initialized = true;
    strace_mode = false;
    silent_mode = false;
#ifdef STRACE
    strace_mode = true;
#endif
    initappspecific();
    this->mc = mc;
    app.mm.init(hp, sp);
    for (int i = 0; sc_list[i]; i += 3) {
        int index = strtoul(sc_list[i], NULL, 10);
        sc_name[index] = sc_list[i + 1];
        sc_arg[index]  = sc_list[i + 2];
    }
    if (strace_mode) {
        printf("## - %s %s\n", L_NAME_SYSCALL, L_VER_SYSCALL);
        printf("##   heap pointer = %08x\n", app.mm.heapptr);
        printf("##   mmap pointer = %08x\n", app.mm.mmapptr);
    }
}

/**********************************************************************/
void SystemCallEmulator::syscall (uint032_t *r, uint064_t cycle)  
{
    kind = r[REG_V0];
    std::string name    = sc_name[kind];
    std::string argkind = sc_arg[kind];
    a[0] = r[REG_A0];
    a[1] = r[REG_A1];
    a[2] = r[REG_A2];
    a[3] = r[REG_A3];
    if (argkind.length() > 4) {
        memread(r[REG_SP] + 16, 4, (uint032_t *) &a[4]);
        memread(r[REG_SP] + 20, 4, (uint032_t *) &a[5]);
    }
    for (int i = 0; i < 2; i++) {
        if (argkind[i] == 's') {
            uint008_t str[NUM_FILENAME_BYTES];
            memstrcpy(str, a[i]);
            a_str[i] = (char *)str;
        }
    }
    if (argkind[0] == 'f') {
        a_fd = getfd(a[0]);
    }
    int result = syscall_internal(r, cycle);
    if (result != -1) {
        r[REG_V0] = result;
        r[REG_A3] = 0;
    } else {
        r[REG_V0] = syscall_errno;
        r[REG_A3] = 1;
    }
    puttrace(r);
}

/**********************************************************************/
int SystemCallEmulator::syscall_internal(uint032_t *r, uint064_t cycle)  
{
    std::string name    = sc_name[kind];
    std::string argkind = sc_arg[kind];
    std::vector<uint032_t> logvec;
    logvec.push_back(kind);
    int result = 0;
    if (argkind[0] == 's')
        setcwd();

    if (name == "open" || name == "creat") {    /**********************/
        int flag = 0, mode = 0;
        if (name == "open") {
            for (int i = 0; sc_conv_open[i] != -1; i += 2) {
                if (a[1] & sc_conv_open[i])
                    flag |= sc_conv_open[i + 1];
            }
            mode = a[2];
        } else {
            flag = O_CREAT | O_WRONLY | O_TRUNC;
            mode = a[1];
        }
        result = open((char *) a_str[0].c_str(), flag, mode);
        syscall_errno = errno;
        app.fdfd[result] = result;
        logvec.push_back(flag);
        logvec.push_back(mode);
        logvec.push_back(result);
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "close") {               /**********************/
        if (a[0] >= 3) {
            result = close(a_fd);
            syscall_errno = errno;
        }
        addlog(2, kind, a[0]);
    } else if (name == "read") {                /**********************/
        uint008_t tmp[a[2]];
        result = read(a[0], tmp, a[2]);
        syscall_errno = errno;
        memmemcpy(a[1], tmp, a[2]);
        addlog(3, kind, a[0], a[2]);
    } else if (name == "write") {               /**********************/
        uint008_t tmp[a[2]];
        memmemcpy(tmp, a[1], a[2]);
        result = a[2];
        if (! silent_mode) {
            if (a_fd == STDOUT_FILENO) {
                std::cout.write((char *) tmp, a[2]);
                std::cout.flush();
            } else if (a_fd == STDERR_FILENO) {
                std::cerr.write((char *) tmp, a[2]);
                std::cerr.flush();
            } else {
                result = write(a_fd, tmp, a[2]);
                syscall_errno = errno;
            }
        }
        logvec.push_back(a[0]);
        for (int i = 0; i < a[2]; i++)
            logvec.push_back(tmp[i]);
        addlog(logvec);
    } else if (name == "writev") {              /**********************/
        logvec.push_back(a[0]);
        for (int i = 0; i < a[2]; i++) {
            uint032_t addr, length;
            memread(a[1] + i * 8 + 0, 4, &addr);
            memread(a[1] + i * 8 + 4, 4, &length);
            uint008_t tmp[length];
            memmemcpy(tmp, addr, length);
            int tmp_result = write(a_fd, tmp, length);
            if (tmp_result == (int) length) {
                result += tmp_result;
            } else {
                syscall_errno = errno;
            }
            for (uint i = 0; i < length; i++)
                logvec.push_back(tmp[i]);
        }
        addlog(logvec);
    } else if (name == "lseek") {               /**********************/
        result = (int) lseek(a_fd, (off_t) a[1], a[2]);
        syscall_errno = errno;
        addlog(4, kind, a[0], a[1], a[2]);
    } else if (name == "ftruncate") {           /**********************/
        result = ftruncate(a_fd, a[1]);
        syscall_errno = errno;
        addlog(3, kind, a[0], a[1]);
    } else if (name == "dup") {                 /**********************/
        result = dup(a_fd);
        syscall_errno = errno;
        app.fdfd[result] = result;
        addlog(3, kind, a[0], result);
    } else if (name == "dup2") {                /**********************/
        int newfd = dup(a_fd);
        result = (newfd != -1) ? a[1] : -1;
        app.fdfd[result] = newfd;
        addlog(4, kind, a[0], a[1], result);
    } else if (name == "ioctl") {               /**********************/
#ifndef SCPLUS_NO_IOCTL
        int req = 0;
        uint008_t tmp[NUM_IOCTL_BYTES];
        for (int i = 0; sc_conv_ioctl[i] != -1; i += 2) {
            if (a[1] == sc_conv_ioctl[i])
                req = sc_conv_ioctl[i + 1];
        }
        if (req) {
            memmemcpy(tmp, a[2], NUM_IOCTL_BYTES);
            result = ioctl(a_fd, req, tmp);
            syscall_errno = errno;
            memmemcpy(a[2], tmp, NUM_IOCTL_BYTES);
            logvec.push_back(a[0]);
            logvec.push_back(req);
            for (uint i = 0; i < NUM_IOCTL_BYTES; i++)
                logvec.push_back(tmp[i]);
            addlog(logvec);
        }
#endif
    } else if (name == "fcntl") {               /**********************/
        result = fcntl(a_fd, a[1], a[2]);
        syscall_errno = errno;
        addlog(4, kind, a[0], a[1], a[2]);
    } else if (name == "pipe") {                /**********************/
        int fd[2];
        if (pipe2(fd, O_NONBLOCK) == 0) {
            result = fd[0];
            r[3] = fd[1];
            app.fdfd[fd[0]] = fd[0];
            app.fdfd[fd[1]] = fd[1];
        } else {
            result = -1;
        }
        addlog(3, kind, fd[0], fd[1]);
    } else if (name == "unlink") {              /**********************/
        result = unlink((char *) a_str[0].c_str());
        syscall_errno = errno;
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "getcwd") {              /**********************/
        uint008_t tmp[a[1]];
        result = (getcwd((char *) tmp, a[1]) != NULL) ? 0 : -1;
        syscall_errno = errno;
        memmemcpy(a[0], (char *) tmp, strlen((char *) tmp) + 1);
    } else if (name == "chdir") {               /**********************/
        result = chdir((char *) a_str[0].c_str());
        syscall_errno = errno;
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "chmod") {               /**********************/
        result = chmod((char *) a_str[0].c_str(), a[1]);
        syscall_errno = errno;
        logvec.push_back(a[1]);
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "chown") {               /**********************/
        result = chown((char *) a_str[0].c_str(), a[1], a[2]);
        syscall_errno = errno;
        logvec.push_back(a[1]);
        logvec.push_back(a[2]);
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "access") {              /**********************/
        result = access((char *) a_str[0].c_str(), a[1]);
        syscall_errno = errno;
    } else if (name == "stat" ||                /**********************/
               name == "lstat" ||
               name == "fstat") {
        uint032_t tmp[35];
        struct stat host_stat;
        if (name == "stat") {
            result = stat((char *) a_str[0].c_str(), &host_stat);
        } else if (name == "lstat") {
            result = lstat((char *) a_str[0].c_str(), &host_stat);
        } else {
            result = fstat(a_fd, &host_stat);
        }
        syscall_errno = errno;

        // convert
        for (int i = 0; i < 35; i++)
            tmp[i] = 0;
        tmp[ 0] = host_stat.st_dev;
        tmp[ 4] = host_stat.st_ino;
        tmp[ 5] = host_stat.st_mode;
        tmp[ 6] = host_stat.st_nlink;
        tmp[ 7] = host_stat.st_uid;
        tmp[ 8] = host_stat.st_gid;
        tmp[ 9] = host_stat.st_rdev;
        tmp[12] = host_stat.st_size;
        tmp[14] = host_stat.st_atime;
        tmp[16] = host_stat.st_mtime;
        tmp[18] = host_stat.st_ctime;
        tmp[20] = host_stat.st_blksize;
        tmp[21] = host_stat.st_blocks;
        memmemcpy(a[1], tmp, sizeof(tmp));
    } else if (name == "utime") {               /**********************/
        int tbuf[2] = {0, 0};
        if (a[1] == 0) {
            result = utime((const char *) a_str[0].c_str(), NULL);
        } else {
            memmemcpy(tbuf, a[1], sizeof(tbuf));
            struct utimbuf ut;
            ut.actime  = tbuf[0];
            ut.modtime = tbuf[1];
            result = utime((const char *) a_str[0].c_str(), &ut);
        }
        syscall_errno = errno;
        logvec.push_back(tbuf[0]);
        logvec.push_back(tbuf[1]);
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "getdents") {            /**********************/
#ifndef SCPLUS_NO_GETDENTS
        uint008_t tmp[a[2]];
        int dents_pos = 0, dents_length;
        dents_length = ::syscall(__NR_getdents, a[0], tmp, a[2]);
        syscall_errno = errno;
        if (sizeof(off_t) == 8) { // convert (only if host is 64-bit)
            while (dents_pos < dents_length) {
                for (int i = dents_pos + 4; i < dents_pos + 8; i++)
                    tmp[i] = tmp[i + 4];
                for (int i = dents_pos + 8; i < a[2] - 8; i++)
                    tmp[i] = tmp[i + 8];
                short *reclen = (short *) &tmp[dents_pos + 8];
                *reclen -= 8;
                dents_pos += *reclen;
                dents_length -= 8;
            }
        }
        result = dents_length;
        memmemcpy(a[1], tmp, dents_length);
        addlog(3, kind, a[0], a[2]);
#endif
    } else if (name == "rename") {              /**********************/
        result = rename((char *) a_str[0].c_str(),
                        (char *) a_str[1].c_str());
        syscall_errno = errno;
        pushstrtolog(logvec, a_str[0]);
        logvec.push_back(0);
        pushstrtolog(logvec, a_str[1]);
        addlog(logvec);
    } else if (name == "mkdir") {               /**********************/
        result = mkdir((char *) a_str[0].c_str(), a[1]);
        syscall_errno = errno;
        logvec.push_back(a[1]);
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "rmdir") {               /**********************/
        result = rmdir((char *) a_str[0].c_str());
        syscall_errno = errno;
        pushstrtolog(logvec, a_str[0]);
        addlog(logvec);
    } else if (name == "brk") {                 /**********************/
        result = app.mm.extendheap(a[0]);
        if (result == 0) {
            result = -1;
            syscall_errno = ENOMEM;
        }
        addlog(3, kind, result, r[REG_SP]);
    } else if (name == "mmap") {                /**********************/
        result = app.mm.findmap(a[0], a[1], r[REG_SP]);
        if (result == 0) {
            result = -1;
            syscall_errno = ENOMEM;
        } else if (! (a[3] & 0x800)) {
            uint008_t tmp[a[1] + a[5]];
            uint readcnt = 0;
            readcnt += read((uint032_t) a[4], tmp, a[1] + a[5]);
            memmemcpy(result, &tmp[a[5]], a[1]);
        }
        addlog(4, kind, a[0], a[1], r[REG_SP]);
    } else if (name == "munmap") {              /**********************/
        app.mm.remap(a[0], a[1], 0);
        addlog(3, kind, a[0], a[1]);
    } else if (name == "mremap") {              /**********************/
        if (! app.mm.remap(a[0], a[1], a[2])) {
            result = -1;
            syscall_errno = ENOMEM;
            fprintf(stderr, "## mremap(): can't expand (old: %d new: %d).\n",
                    a[1], a[2]);
        }
        addlog(4, kind, a[0], a[1], a[2]);
    } else if (name == "getuid") {              /**********************/
        app.uid  = (app.uid > 0)  ? app.uid  : getuid();
        app.euid = (app.euid > 0) ? app.euid : geteuid();
        result   = (app.uid > 0)  ? app.uid  : -1;
        syscall_errno = errno;
        addlog(3, kind, result, app.euid);
    } else if (name == "geteuid") {             /**********************/
        app.euid = (app.euid > 0) ? app.euid : geteuid();
        result   = (app.euid > 0) ? app.euid : -1;
        syscall_errno = errno;
        addlog(2, kind, result);
    } else if (name == "getgid") {              /**********************/
        app.gid  = (app.gid > 0)  ? app.gid  : getgid();
        app.egid = (app.egid > 0) ? app.egid : getegid();
        result   = (app.egid > 0) ? app.egid : -1;
        syscall_errno = errno;
        addlog(3, kind, result, app.egid);
    } else if (name == "getegid") {             /**********************/
        app.egid = (app.egid > 0) ? app.egid : getegid();
        result   = (app.egid > 0) ? app.egid : -1;
        syscall_errno = errno;
        addlog(2, kind, result);
    } else if (name == "getpid") {              /**********************/
        app.pid  = (app.pid > 0)  ? app.pid  : getpid();
        result   = (app.pid > 0)  ? app.pid  : -1;
        syscall_errno = errno;
        addlog(2, kind, result);
    } else if (name == "getrlimit" ||           /**********************/
               name == "setrlimit") {
#ifndef SCPLUS_NO_RESOURCE
        uint032_t tmp[2];
        struct rlimit rl;
        memmemcpy(tmp, a[1], sizeof(tmp));
        rl.rlim_cur = tmp[0];
        rl.rlim_max = tmp[1];
        if (name == "getrlimit")
            result = getrlimit(a[0], &rl);
        else
            result = setrlimit(a[0], &rl);
        syscall_errno = errno;
        tmp[0] = rl.rlim_cur;
        tmp[1] = rl.rlim_max;
        memmemcpy(a[1], tmp, sizeof(tmp));
#endif
    } else if (name == "getrusage") {           /**********************/
#ifndef SCPLUS_NO_RESOURCE
        uint032_t tmp[18];
        struct rusage ru;
        result = getrusage(a[0], &ru);
        syscall_errno = errno;
        tmp[ 0] = ru.ru_utime.tv_sec;
        tmp[ 1] = ru.ru_utime.tv_usec;
        tmp[ 2] = ru.ru_stime.tv_sec;
        tmp[ 3] = ru.ru_stime.tv_usec;
        tmp[ 4] = ru.ru_maxrss;
        tmp[ 5] = ru.ru_ixrss;
        tmp[ 6] = ru.ru_idrss;
        tmp[ 7] = ru.ru_isrss;
        tmp[ 8] = ru.ru_minflt;
        tmp[ 9] = ru.ru_majflt;
        tmp[10] = ru.ru_nswap;
        tmp[11] = ru.ru_inblock;
        tmp[12] = ru.ru_oublock;
        tmp[13] = ru.ru_msgsnd;
        tmp[14] = ru.ru_msgrcv;
        tmp[15] = ru.ru_nsignals;
        tmp[16] = ru.ru_nvcsw;
        tmp[17] = ru.ru_nivcsw; 
        memmemcpy(a[1], tmp, sizeof(tmp));
#endif
    } else if (name == "time" ||                /**********************/
               name == "gettimeofday" ||
               name == "clock_gettime") {
        struct timeval tv;
        result = gettimeofday(&tv, NULL);
        syscall_errno = errno;
        if (app.basecycle == 0) {
            app.basetime  = (uint064_t) tv.tv_sec * 1000000 + tv.tv_usec;
            app.basecycle = cycle;
        }
        uint064_t ctime = cycle + app.startcycle;
        if (name != "clock_gettime" || a[0] == 0) {
            ctime = (app.basetime + 
                     (ctime - app.basecycle) / (FREQUENCY / 1000000));
        } else {
            ctime = ctime / (FREQUENCY / 1000000);
        }
        uint032_t tmp[2] = {ctime / 1000000, ctime % 1000000};
        if (name == "time") {
            result = tmp[0];
            if (a[0])
                memmemcpy(a[0], tmp, 4);
            addlog(4, kind, app.basetime / 1000000,
                   (uint032_t) (cycle >> 32), (uint032_t) cycle);
        } else if (name == "gettimeofday") {
            if (a[0])
                memmemcpy(a[0], tmp, 8);
            addlog(5, kind, 
                   (uint032_t) (app.basetime >> 32), (uint032_t) app.basetime,
                   (uint032_t) (cycle >> 32), (uint032_t) cycle);
        } else { // clock_gettime
            tmp[1] *= 1000; // usec -> nsec
            if (a[1])
                memmemcpy(a[1], tmp, 8);
            addlog(3, kind, 
                   (uint032_t) (cycle >> 32), (uint032_t) cycle);
        }
    } else if (name == "times") {               /**********************/
#ifndef SCPLUS_NO_TIMES
        struct tms ts;
        result = (uint032_t) times(&ts);
        syscall_errno = errno;
        if (app.basecycle_tms == 0) {
            app.basecycle_tms = cycle;
            app.basetms       = ts;
        }
        uint064_t inc = ((cycle + app.startcycle - app.basecycle_tms)
                         * 100 / FREQUENCY);
        uint032_t tmp[4];
        tmp[0] = app.basetms.tms_utime + inc;
        tmp[1] = app.basetms.tms_stime + inc;
        tmp[2] = app.basetms.tms_cutime + inc;
        tmp[3] = app.basetms.tms_cstime + inc;
        if (a[0])
            memmemcpy(a[0], tmp, sizeof(tmp));
        addlog(7, kind,
               app.basetms.tms_utime,  app.basetms.tms_stime,
               app.basetms.tms_cutime, app.basetms.tms_cstime,
               (uint032_t) (cycle >> 32), (uint032_t) cycle);
#endif
    } else {                                    /**********************/
        printf("!! unknown syscall #%d. Skip.\n", kind);
    }

    if (argkind[0] == 's')
        restorecwd();
    return result;
}

/**********************************************************************/
void SystemCallEmulator::initappspecific()  
{
    app.startcycle = app.basecycle = app.basecycle_tms = app.basetime = 0;
    app.pid = app.uid = app.euid = app.gid = app.egid = 0;
#ifndef SCPLUS_NO_TIMES
    app.basetms.tms_utime = 0;
    app.basetms.tms_stime = 0;
    app.basetms.tms_cutime = 0;
    app.basetms.tms_cstime = 0;
#endif
}

/**********************************************************************/
bool SystemCallEmulator::memstrcpy(uint008_t *str, uint064_t addr)  
{
    for (uint i = 0; i < NUM_FILENAME_BYTES; i++) {
        memread(addr + i, 1, &str[i]);
        if (str[i] == '\0')
            return true;
    }
    fprintf(stderr, "!! SystemCallEmulator: too long filename!\n");
    exit(1);
}

/**********************************************************************/
bool SystemCallEmulator::memmemcpy(void *buf, uint064_t addr, uint size)  
{
    uint008_t *str = (uint008_t *) buf;
    for (uint i = 0; i < size; i++)
        memread(addr + i, 1, &str[i]);
    return true;
}

/**********************************************************************/
bool SystemCallEmulator::memmemcpy(uint064_t addr, void *buf, uint size)  
{
    uint008_t *str = (uint008_t *) buf;
    for (uint i = 0; i < size; i++)
        memwrite(addr + i, 1, (uint064_t) str[i]);
    return true;
}

/**********************************************************************/
int SystemCallEmulator::getfd(int fd)
{
    if (app.fdfd.find(fd) != app.fdfd.end())
        return app.fdfd[fd];
    return fd;
}

/**********************************************************************/
void SystemCallEmulator::setcwd()  
{
    char currentdir[PATH_MAX + 1];
    if (getcwd(currentdir, PATH_MAX) != NULL)
        sim_cwd = currentdir;
    int x = chdir(app.cwd.c_str());
    x = x;
}

/**********************************************************************/
void SystemCallEmulator::updatecwd()  
{
    char currentdir[PATH_MAX + 1];
    if (getcwd(currentdir, PATH_MAX) != NULL)
        app.cwd = currentdir;
}

/**********************************************************************/
void SystemCallEmulator::addlog(int num, ...)
{
    va_list list;
    std::vector<uint032_t> newvec;
    va_start(list, num);
    for (int i = 0; i < num; i++)
        newvec.push_back(va_arg(list, uint032_t));
    addlog(newvec);
    va_end(list);
}

/**********************************************************************/
void SystemCallEmulator::puttrace(uint032_t *r)  
{
    if (!strace_mode)
        return;
    std::string name   = sc_name[kind];
    std::string argkind = sc_arg[kind];
    printf("## %s(", name.c_str());
    for (uint i = 0; i < argkind.length(); i++) {
        if (argkind[i] == 'i')
            printf("%d", a[i]);
        else if (argkind[i] == 'u')
            printf("0x%x", a[i]);
        else if (argkind[i] == 's')
            printf("0x%x<%s>", a[i], a_str[i].c_str());
        else if (argkind[i] == 'f')
            printf("%d<%d>", a[i], a_fd);
        if (i == argkind.length() - 1)
            printf(")=");
        else
            printf(", ");
    }
    if (r[REG_A3] == 0)
        printf("%d\n", r[REG_V0]);
    else
        printf("%d(%s)\n", r[REG_V0], strerror(r[REG_V0]));
}

/* HeapAllocator ******************************************************/
/**********************************************************************/
void HeapAllocator::init(uint032_t hp, uint032_t sp)  
{
    this->heapptr = ROUND_UP(hp);
    uint032_t heap_size = (sp - hp) / 4;
    uint032_t mmap_size = sp - hp - heap_size;
    this->mmapptr = ROUND_UP(hp + heap_size);
    table.assign(mmap_size / MM_PAGE_SIZE + 1, 0);
}

/**********************************************************************/
uint032_t HeapAllocator::extendheap(uint032_t addr)  
{
    addr = ROUND_UP(addr);
    if (addr < mmapptr) {
        heapptr = (heapptr < addr) ? addr : heapptr;
        return heapptr;
    }
    return 0;
}

/**********************************************************************/
uint032_t HeapAllocator::findmap(uint032_t addr, uint032_t size, uint032_t sp)  
{
    if ((sp - mmapptr) / MM_PAGE_SIZE + 1 < (uint) table.size())
        table.resize((sp - mmapptr) / MM_PAGE_SIZE + 1);
    addr = ROUND_UP(addr);
    addr = (addr > mmapptr) ? addr - mmapptr : 0;
    uint index = addr / MM_PAGE_SIZE;
    uint pages = (size + MM_PAGE_SIZE - 1) / MM_PAGE_SIZE;
    int sum = 0;
    for (uint i = index; i < index + pages; i++)
        sum += table[i];
    for (uint i = index; i < table.size() - pages; i++) {
        if (sum == 0) {
            for (uint j = i; j < i + pages; j++)
                table[j] = i + 1;
            return mmapptr + i * MM_PAGE_SIZE;
        }
        sum += table[i + pages] - table[i];
    }
    return 0;
}

/**********************************************************************/
bool HeapAllocator::remap(uint032_t addr,
                          uint032_t oldsize, uint032_t newsize)  
{
    if (addr < mmapptr)
        return false;
    addr = (addr - mmapptr) / MM_PAGE_SIZE;
    if (newsize == 0) {
        for (uint i = addr; table[i] == addr + 1; i++)
            table[i] = 0;
    } else if (oldsize < newsize) {
        newsize = (newsize + MM_PAGE_SIZE - 1) / MM_PAGE_SIZE;
        for (uint i = addr; i < addr + newsize; i++) {
            if (i >= table.size())
                return false;
            if (table[i] != 0 && table[i] != addr + 1)
                return false;
            table[i] = addr + 1;
        }
    }
    return true;
}

/* System Call Definition *********************************************/
/**********************************************************************/
const char *SystemCallEmulator::sc_list[] = {
    "4001", "exit"         , "",
    "4005", "open"         , "sii",
    "4006", "close"        , "f",
    "4003", "read"         , "fui",
    "4004", "write"        , "fui",
    "4146", "writev"       , "fui",
    "4019", "lseek"        , "fui",
    "4093", "ftruncate"    , "fi",
    "4041", "dup"          , "f",
    "4063", "dup2"         , "f",
    "4054", "ioctl"        , "fiu",
    "4055", "fcntl"        , "fii",
    "4042", "pipe"         , "",
    "4008", "creat"        , "sii",
    "4010", "unlink"       , "s",
    "4203", "getcwd"       , "ui",
    "4012", "chdir"        , "s",
    "4015", "chmod"        , "si",
    "4202", "chown"        , "sii",
    "4033", "access"       , "si",
    "4106", "stat"         , "su",
    "4107", "lstat"        , "su",
    "4108", "fstat"        , "fu",
    "4030", "utime"        , "su",
    "4141", "getdents"     , "fui",
    "4039", "mkdir"        , "s",
    "4038", "rename"       , "ss",
    "4040", "rmdir"        , "s",
    "4045", "brk"          , "u",
    "4090", "mmap"         , "uuiuiu",
    "4167", "mremap"       , "uii",
    "4091", "munmap"       , "ui",
    "4024", "getuid"       , "",
    "4049", "geteuid"      , "",
    "4047", "getgid"       , "",
    "4050", "getegid"      , "",
    "4020", "getpid"       , "",
    "4075", "setrlimit"    , "ui",
    "4076", "getrlimit"    , "ui",
    "4077", "getrusage"    , "iu",
    "4013", "time"         , "u",
    "4043", "times"        , "u",
    "4078", "gettimeofday" , "uu",
    "4263", "clock_gettime", "uu",
    NULL, NULL, NULL
};

/* Guest-Specific Constant Definition *********************************/
/**********************************************************************/
const int SystemCallEmulator::sc_conv_open[] = {
    0x00000, O_RDONLY,
    0x00001, O_WRONLY,
    0x00002, O_RDWR,
    0x00010, O_SYNC,
    0x00100, O_CREAT,
    0x00200, O_TRUNC,
    0x00400, O_EXCL,
    0x00800, O_NOCTTY,
    0x01000, O_ASYNC,
    0x10000, O_DIRECTORY,
    -1, -1
};

const int SystemCallEmulator::sc_conv_ioctl[] = {
    0x00007408, LOCAL_TIOCGETP,
    0x00007409, LOCAL_TIOCSETP,
    0x00005401, LOCAL_TCGETA,
    0x00007475, LOCAL_TIOCSLTC,
    0x40087468, LOCAL_TIOCGWINSZ,
    0x00005403, LOCAL_TCSETAW,
    -1, -1
};

/**********************************************************************/
