/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "env.h"
#include "version.h"
#include "common/memorymap.h"
#include "network/packet.h"
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <iostream>
#include <sstream>
#include <iomanip>

/**********************************************************************/
volatile sig_atomic_t receive_int = 0;

void sigint_handler(int signum)
{
    signum = signum;
    receive_int = 1;
}

/**********************************************************************/
int main (int argc, char **argv)
{
    Env env;
    std::cout << "## " << MC_NAME << ' ' << MC_VER << std::endl;
    return env.sim_main(argc, argv);
}

/**********************************************************************/
Env::Env()
{
    configfile = "";
    x_rank     = SIZE_X_DEF;
    y_rank     = SIZE_Y_DEF;
    add_server = false;
    add_memory = false;
    add_smart  = false;
    max_cycle  = MAX_CYCLE_DEF;
    num_core   = 0;
}

/**********************************************************************/
void Env::usage()
{   
    static const char *usagemessage = "\n\
usage: ./SimMc [options] binary1 binary2 ...\n\
  --config [file], -c [file]: set a system config file\n\
  --x-rank num   , -x[num]  : number of columns \n\
  --y-rank num   , -y[num]  : number of rows \n\
  --end-cycle num, -e[num]  : stop simulation after executing num cycle \n\
  --server       , -A       : * add a row for server nodes \n\
  --no-server    , -B       : * do not add server nodes (default) \n\
  --memory                  : * add a memory node and path nodes \n\
  --no-memory               : * do not add a memory node (default) \n\
  --(no-)smart              : * enable/disable SmartCore Unit \n\
  --smart-pattern string    : * node allocation pattern for SmartCore\n\
      available patterns are INTERLEAVE, CHECK and BLOCK\n\
  --help                    : show this message\n\
  --version                 : show version\n\
  * marked options are invalidated if --config is set \n\
sample: \n\
  SimMc -x8 -y8 test.out \n\
";
    std::cout << usagemessage;
}

/**********************************************************************/
ullint Env::gettime()
{
    static ullint start = 0;
    ullint tm;
    struct timeval tv;
    
    gettimeofday(&tv, NULL);
    tm = (ullint)tv.tv_sec * 1000000ull + tv.tv_usec;
    if (start == 0)
        start = tm;
    return tm - start;
}

/**********************************************************************/
double Env::readargnum (char *buf)
{
    if (buf == NULL)
        return -1;
    std::istringstream iss(buf);
    double d = 0;
    iss >> d;
    return d;
}

/**********************************************************************/
bool Env::readargs (int argc, char **argv)
{
    int index = 1;
    std::cout << "## arguments:";
    while (index < argc) {
        int headindex = index;
        std::cout << ' ' << argv[index];
        if (argv[index][0] == '-') {
            if (argv[index][1] == 'x') {
                x_rank     = (int) readargnum(&argv[index][2]);
            } else if (argv[index][1] == 'y') {
                y_rank     = (int) readargnum(&argv[index][2]);
            } else if (argv[index][1] == 'c') {
                if (argv[++index])
                    configfile = argv[index];
            } else if (argv[index][1] == 'e') {
                max_cycle  = (int) readargnum(&argv[index][2]);
            } else if (argv[index][1] == 'A') {
                add_server = true;
            } else if (argv[index][1] == 'B') {
                add_server = false;
            } else if (argv[index][1] == '-') {
                std::string argname(&argv[index][2]);
                bool value = true;
                if (argname.find("no-", 0) == 0) {
                    argname.erase(0, 3);
                    value = false;
                }
                if      (argname == "x-rank")
                    x_rank     = (int)    readargnum(argv[++index]);
                else if (argname == "y-rank")
                    y_rank     = (int)    readargnum(argv[++index]);
                else if (argname == "end-cycle")
                    max_cycle  = (ullint) readargnum(argv[++index]);
                else if (argname == "server")
                    add_server = value;
                else if (argname == "memory")
                    add_memory = value;
                else if (argname == "smart")
                    add_smart  = value;
                else if (argname == "smart-pattern") {
                    if (argv[++index])
                        smart_pattern = argv[index];
                } else if (argname == "config") {
                    if (argv[++index])
                        configfile = argv[index];
                } else if (argname == "help") {
                    std::cout << std::endl;
                    usage();
                    exit(0);
                } else if (argname == "version") {
                    std::cout << std::endl;
                    exit(0);
                } else {
                    binaryfiles.push_back(argv[index]);
                }
            } else {
                binaryfiles.push_back(argv[index]);
            }
        } else {
            binaryfiles.push_back(argv[index]);
        }
        if (headindex != index && argv[index])
            std::cout << ' ' << argv[index];
        index++;
    }
    std::cout << std::endl;
    return true;
}

/**********************************************************************/
void Env::writeapp(MipsFP &c, MainMemory &m, SimLoader &ld,
                   uint032_t heapptr, uint032_t globalptr, std::string name)
{
    for (int i = 0; i < ld.memtabnum; i++)
        m.write(ld.memtab[i].addr, 1, ld.memtab[i].data);
    for (uint i = 0; i < name.length(); i++)
        m.write(SIZE_NODE_MEMORY_DEF - 0x1000 + i, 1, name[i]);
    m.write(SIZE_NODE_MEMORY_DEF - 0x2000    , 4, 1);
    m.write(SIZE_NODE_MEMORY_DEF - 0x2000 + 4, 4,
            SIZE_NODE_MEMORY_DEF - 0x1000);

    for (int i = 0; i < NREG; i++)
        c.as->r[i] = 0;
    for (int i = 0; i < NFREG; i++)
        c.asfp->fr.w[i] = 0;
    c.as->r[REG_SP] = SIZE_NODE_MEMORY_DEF - 0x2000;
    c.as->pc = c.as->r[REG_T9] = ld.entry;
    c.ss->heapptr   = heapptr;
    c.as->r[REG_GP] = globalptr;
    c.state = CPU_START;
}

/**********************************************************************/
bool Env::init(int argc, char **argv)
{
    std::vector<SimLoader> ld;
    std::vector<uint032_t> heapptr, globalptr;

    // default settings
    sc.num_channel  = NUM_CHANNEL_DEF;
    sc.num_inbuf    = BUFFER_LENGTH_DEF;
    sc.lookahead_rc = true;
    sc.fix_vc       = true;
    sc.fix_priority = false;
    sc.credit_fc    = true;
    sc.skip_lt      = false;    

    // read & check arguments
    if (argc == 1)
        return false; // just print usage
    readargs(argc, argv);
    if (configfile == "") {
        if (x_rank <= 0 || y_rank <= 0 || x_rank * y_rank <= 0) {
            std::cerr << "!! invalid physical size" << std::endl;
            return false;
        }
        sc.setdefaultconfig(x_rank, y_rank, add_memory, add_server,
                            add_smart, smart_pattern);
    } else {
        if (! sc.readconfigfile(configfile.c_str())) {
            std::cerr << "!! failed to parse config file" << std::endl;
            return false;
        }            
    }
    x_rank = sc.x_rank;
    y_rank = sc.y_rank;
    if ((int) binaryfiles.size() != sc.num_app) {
        std::cerr << "!! the numbers of application(s) do not match" 
                  << std::endl;
        std::cerr << "!!  system config = " << sc.num_app << std::endl;
        std::cerr << "!!  arguments     = " << binaryfiles.size()
                  << std::endl;
        for (uint i = 0; i < binaryfiles.size(); i++)
            std::cerr << "!!   [" << (i + 1) << "] " 
                      << binaryfiles[i] << std::endl;
        return false;
    }

    // load binary files
    ld.resize       (sc.num_app);
    heapptr.resize  (sc.num_app);
    globalptr.resize(sc.num_app);
    for (int i = 0; i < sc.num_app; i++) {
        if (ld[i].loadfile(binaryfiles[i].c_str())) {
            std::cerr << "!! failed to load binary file: " 
                      << binaryfiles[i] << std::endl;
            return false;
        }
        for (int j = 0; j < ld[i].memtabnum; j++)
            if (heapptr[i] < ld[i].memtab[j].addr)
                heapptr[i] = ld[i].memtab[j].addr;
        for (int j = 0; j < ld[i].symtabnum; j++)
            if (strcmp(ld[i].symtab[j].name, "_gp") == 0)
                globalptr[i] = ld[i].symtab[j].addr;
    }
    sc.max_packet_length = Packet::MAX_FLIT;
    sc.print();

    // initialize and connect chip elements
    std::vector<nodetype_t> nodetype;
    std::vector<int>        smart_on;
    int num_smart = 0, num_server = 0, num_memory = 0;
    router.resize  (x_rank * y_rank);
    nodetype.resize(x_rank * y_rank);
    smart_on.resize(x_rank * y_rank, 0);
    for (uint i = 0; i < x_rank * y_rank; i++) {
        // 1st loop: init routers & specify node type
        uint x    = i % x_rank + 1;
        uint y    = i / x_rank + 1;
        uint pid  = ID(x, y);
        router[i].init(pid, sc);
        if (! sc.isvalid(pid)) {
            nodetype[i] = PATH;
        } else if (sc.getlid(pid) == ID(0, 0)) {
            nodetype[i] = MEMORY;
            num_memory++;
        } else {
            uint rank = sc.getlrank(sc.getaid(pid));
            if (IDX(sc.getlid(pid)) > IDX(rank) ||
                IDY(sc.getlid(pid)) > IDY(rank)) {
                nodetype[i] = SERVER;
                num_server++;
            } else { 
                nodetype[i] = COMPUTATION;
                num_core++;
            }
        }
        if (sc.isvalid(pid) && (sc.ismaster(pid) || sc.isslave(pid))) {
            smart_on[i] = 1;
            num_smart++;
        }
    }
    core.resize(num_core + num_server);
    mem.resize (num_core + num_server + num_memory);
    incc.resize(num_core + num_server + num_memory);
    smart.resize(num_smart);
    int index_core   = 0;
    int index_server = num_core;
    int index_memory = num_core + num_server;
    int index_smart  = 0;
    for (uint i = 0; i < x_rank * y_rank; i++) {
        // 2nd loop: connect routers & init nodes
        uint x    = i % x_rank + 1;
        uint y    = i / x_rank + 1;
        uint pid  = ID(x, y);
        int index = 0;
        if (x != x_rank) {
            router[i].connect(DIR_E, router[i + 1], DIR_W);
            router[i + 1].connect(DIR_W, router[i], DIR_E);
        }
        if (y != y_rank) {
            router[i].connect(DIR_S, router[i + sc.x_rank], DIR_N);
            router[i + sc.x_rank].connect(DIR_N, router[i], DIR_S);            
        }
        if (nodetype[i] == MEMORY) {
            index = index_memory++;
            mem[index].init(SIZE_MAIN_MEMORY_DEF);
        } else if (nodetype[i] != PATH) {
            int  app  = sc.getaid(pid);
            if (nodetype[i] == COMPUTATION)
                index = index_core++;
            else
                index = index_server++;
            mem[index].init(SIZE_NODE_MEMORY_DEF);
            core[index].init(&mem[index]);
            core[index].setpidstr(x, y);
            writeapp(core[index], mem[index], ld[app],
                     heapptr[app], globalptr[app], binaryfiles[app]);
            if (sc.isslave(pid))
                core[index].ss->silent_mode = true;
        }
        if (nodetype[i] != PATH) {
            if (! smart_on[i]) {
                incc[index].init(pid, sc, &mem[index]);
                incc[index].connect(0, router[i], DIR_C);
                router[i].connect(DIR_C, incc[index], 0);
            } else {
                int sindex = index_smart++;
                incc[index].init(pid, sc, &mem[index]);
                smart[sindex].init(pid, sc, &router[i]);
                incc[index].connect(0, smart[sindex], SmartUnit::SIDE_INCC);
                router[i].connect(DIR_C, smart[sindex], SmartUnit::SIDE_RT);
                smart[sindex].connect(SmartUnit::SIDE_INCC, incc[index], 0);
                smart[sindex].connect(SmartUnit::SIDE_RT, router[i], DIR_C);
            }
        }
    }

    // set up signal handler
    struct sigaction sa;
    sa.sa_handler = sigint_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGINT, &sa, NULL);

    return true;
}

/**********************************************************************/
void Env::sim()
{
    std::vector<MipsFP>::iterator     core_it;
    std::vector<MainMemory>::iterator mem_it;
    std::vector<Incc>::iterator       incc_it;
    std::vector<Router>::iterator     rt_it;
    std::vector<SmartUnit>::iterator  smart_it;
    cycle = 0;
    gettime();
    while (running()) {
        for (core_it = core.begin(); core_it != core.end(); core_it++)
            if (core_it->running())
                core_it->drive();
        for (mem_it  = mem.begin() ; mem_it  != mem.end();  mem_it++)
            mem_it->drive();
        for (incc_it = incc.begin(); incc_it != incc.end(); incc_it++)
            incc_it->drive();
        for (rt_it = router.begin(); rt_it != router.end(); rt_it++)
            rt_it->drive();
        for (smart_it = smart.begin(); smart_it != smart.end(); smart_it++)
            smart_it->drive();

        for (core_it = core.begin(); core_it != core.end(); core_it++)
            if (core_it->running())
                core_it->update();
        for (mem_it  = mem.begin() ; mem_it  != mem.end();  mem_it++)
            mem_it->update();
        for (incc_it = incc.begin(); incc_it != incc.end(); incc_it++)
            incc_it->update();
        for (rt_it = router.begin(); rt_it != router.end(); rt_it++)
            rt_it->update();
        for (smart_it = smart.begin(); smart_it != smart.end(); smart_it++)
            smart_it->update();
        cycle++;
    }
    double simtime = (double) gettime() / 1000000;
    std::cout << "## Simulation time (s)    : " << std::fixed
              << std::setprecision(3) << simtime << std::endl;
    std::cout << "## Simulation cycle       : " << cycle << std::endl;
    std::cout << "## Simulation speed (kc/s): " << std::setprecision(3)
              << ((double) cycle / simtime / 1000) << std::endl;
}

/**********************************************************************/
bool Env::running()
{
    if (receive_int)
        return false;
    if (cycle >= max_cycle)
        return false;
    for (int i = 0; i < num_core; i++)
        if (core[i].running())
            return true;
    for (int i = 0; i < sc.x_rank * sc.y_rank; i++)
        if (router[i].isbusy())
            return true;
    return false;
}

/**********************************************************************/
int Env::sim_main(int argc, char **argv)
{
    if (! init(argc, argv)) {
        usage();
        return 1;
    }
    sim();
    return 0;
}

/**********************************************************************/
