/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "incc.h"
#include "../common/id.h"
#include "../common/memorymap.h"
#include "routecomp.h"

/* Incc                                                               */
/**********************************************************************/
void Incc::init(uint pid, SystemConfig &sc, MainMemory *mem) 
{
    Connectable::init(1, sc);
    send.init(pid, sc, mem);
    recv.init(sc, mem);
    this->mem = mem;
    num_channel = sc.num_channel;
    credit_fc   = sc.credit_fc;
    cycle = 0;

    // Memory Map
    if (mem) {
        mem->write(MP_MMAP_LCD        , 4, 1);
        mem->write(MP_MMAP_TIMER      , 4, 0);
        mem->write(SC_MMAP_DMA_PUT_CMD, 4, 0);
        mem->write(SC_MMAP_TIMER      , 4, 0);
        mem->write(SC_MMAP_TIMER_D    , 4, 0);
        mem->write(SC_MMAP_INST_LOOP  , 4, 0x1000ffff); // j (pc+0)
        mem->write(SC_MMAP_INST_LOOP__, 4, 0);
        mem->write(SC_MMAP_PC         , 4, 0);
    }
    if (mem && sc.isvalid(pid)) {
        mem->write(SC_MMAP_NODE_ID    , 4, (pid << 16) | sc.getlid(pid));
        mem->write(SC_MMAP_NOC_SIZE   , 4, sc.getlrank(sc.getaid(pid)));
    }
}

/**********************************************************************/
void Incc::connect(int srcport, Connectable &destunit, int destport)
{
    Connectable::connect(srcport, destunit, destport);
    send.connect(this);
    recv.connect(this);
}

/**********************************************************************/
void Incc::drive()
{
    send.drive();
    recv.drive();
}

/**********************************************************************/
void Incc::update()
{
    send.update();
    recv.update();
    cycle++;
    char lcdchar;
    mem->read (MP_MMAP_LCD        , 1, &lcdchar);
    if (lcdchar != 1) {
        std::cout << lcdchar;
        std::cout.flush();
    }
    mem->write(MP_MMAP_LCD        , 4, 1);
    mem->write(MP_MMAP_TIMER      , 4, cycle / 1000);
    mem->write(SC_MMAP_TIMER      , 4, cycle & 0xffffffff);
    mem->write(SC_MMAP_TIMER_D    , 4, cycle >> 32);
    for (uint i = 0; i < num_channel; i++)
        turnon(0, i); // always accept flits
    if (credit_fc && out_flit[0]->valid)
        deccredit(0, out_flit[0]->outch);
}

/* InccSend                                                           */
/**********************************************************************/
void InccSend::init(uint id, SystemConfig &sc, MainMemory *mem)
{
    this->mem = mem;
    my_id = id;
    my_x = IDX(id);
    my_y = IDY(id);
    num_channel  = sc.num_channel;
    lookahead_rc = sc.lookahead_rc;
    type = count = next_type = next_count = 0;
    deq_get_cmd = deq_ack_cmd = false;

    if (sc.isvalid(id)) {
        laddr_64bit = (sc.getlid(id) == 0);
        lpmap = sc.getlptable(sc.getaid(id));
    } else {
        laddr_64bit = false;
    }
}

/**********************************************************************/
void InccSend::connect(Incc *incc)
{
    this->out_flow = incc->out_flow[0];
    this->out_flit = incc->out_flit[0];
}

/**********************************************************************/
void InccSend::addputcommand()
{
    DmaCommand &cmd = next_put_cmd;
    uint032_t tmpaddr;

    cmd.clear();
    cmd.type = DMA_PUT;
    mem->read(SC_MMAP_DMA_PUT_REMOTE_ID    , 4, &cmd.remote_id);
    mem->read(SC_MMAP_DMA_PUT_REMOTE_ADDR  , 4, &tmpaddr);
    cmd.remote_addr = tmpaddr;
    mem->read(SC_MMAP_DMA_PUT_REMOTE_STRIDE, 4, &cmd.remote_stride);
    mem->read(SC_MMAP_DMA_PUT_LOCAL_ADDR   , 4, &tmpaddr);
    cmd.local_addr  = tmpaddr;
    mem->read(SC_MMAP_DMA_PUT_LOCAL_STRIDE , 4, &cmd.local_stride);
    mem->read(SC_MMAP_DMA_PUT_SIZE         , 4, &cmd.size);
    mem->read(SC_MMAP_DMA_PUT_CH           , 4, &cmd.channel);
    if (cmd.remote_id == ID(0, 0)) { // DMA_PUT64
        cmd.type = DMA_PUT64;
        cmd.remote_addr |= (uint064_t) cmd.remote_stride << 32;
        cmd.remote_stride = 4;
    }
    cmd.remote_id     = cmd.remote_id & 0xffff;
    cmd.local_stride  = (cmd.local_stride == 0)  ? 4 : cmd.local_stride;
    cmd.remote_stride = (cmd.remote_stride == 0) ? 4 : cmd.remote_stride;
}

/**********************************************************************/
void InccSend::addgetcommand()
{
    DmaCommand &cmd = new_get_cmd;
    DmaCommand &ack = new_ack_cmd;
    uint032_t tmpaddr, tmpaddr_h;

    cmd.clear();
    cmd.type = DMA_GET;
    if (! laddr_64bit) { // DMA_GET
        mem->read(SC_MMAP_DMA_GET_REMOTE_ID    , 4, &cmd.remote_id);
        mem->read(SC_MMAP_DMA_GET_REMOTE_ADDR  , 4, &tmpaddr);
        cmd.remote_addr = tmpaddr;
        mem->read(SC_MMAP_DMA_GET_REMOTE_STRIDE, 4, &cmd.remote_stride);
        mem->read(SC_MMAP_DMA_GET_LOCAL_ADDR   , 4, &tmpaddr);
        cmd.local_addr  = tmpaddr;
        mem->read(SC_MMAP_DMA_GET_LOCAL_STRIDE , 4, &cmd.local_stride);
        mem->read(SC_MMAP_DMA_GET_SIZE         , 4, &cmd.size);
    } else {             // DMA_GET64
        mem->read(GMEM_DMA_GET_REMOTE_ID       , 4, &cmd.remote_id);
        mem->read(GMEM_DMA_GET_REMOTE_ADDR     , 4, &tmpaddr);
        cmd.remote_addr = tmpaddr;
        mem->read(GMEM_DMA_GET_REMOTE_STRIDE   , 4, &cmd.remote_stride);
        mem->read(GMEM_DMA_GET_LOCAL_ADDR_L    , 4, &tmpaddr);
        mem->read(GMEM_DMA_GET_LOCAL_ADDR_H    , 4, &tmpaddr_h);
        cmd.local_addr = ((uint064_t) tmpaddr_h << 32) | tmpaddr;
        cmd.local_stride = 4;
        mem->read(GMEM_DMA_GET_SIZE            , 4, &cmd.size);
    }
    mem->read(SC_MMAP_DMA_GET_CH           , 4, &cmd.channel);
    cmd.remote_id     = cmd.remote_id & 0xffff;
    cmd.local_stride  = (cmd.local_stride == 0)  ? 4 : cmd.local_stride;
    cmd.remote_stride = (cmd.remote_stride == 0) ? 4 : cmd.remote_stride;  

    ack.clear();
    ack.type = DMA_ACK;
    ack.remote_id   = cmd.remote_id;
    ack.remote_addr = SC_MMAP_DMA_FLAG;
    ack.remote_stride = ack.local_stride = ack.size = 4;
    ack.channel   = cmd.channel;
    ack.immediate = ((get_cmd.size() != (uint) MAX_ACCEPT_GET) ? 
                     DMAFLAG_ACK : DMAFLAG_NACK);
}

/**********************************************************************/
void InccSend::getnextcommand()
{
    // priority order: PUT > GET > ACK
    if (put_cmd.valid())
        next_cur_cmd = put_cmd;
    else if (! get_cmd.empty())
        next_cur_cmd = get_cmd.front();
    else if (! ack_cmd.empty())
        next_cur_cmd = ack_cmd.front();
    else
        return;
    
    if (next_cur_cmd.type == DMA_PUT64)
        next_type = Packet::PUT64;
    else if (next_cur_cmd.type == DMA_ACK)
        next_type = Packet::PUT32_ACKNACK;
    else if (next_cur_cmd.remote_stride != 4)
        next_type = Packet::PUT32_STRIDE;
    else
        next_type = Packet::PUT32;
}

/**********************************************************************/
void InccSend::generateflit()
{
    out_flit->valid    = true;
    out_flit->head     = (count == 0);
    out_flit->tail     = (Packet::def[type][count + 1][0] == FT_END);
    out_flit->outch    = cur_cmd.channel % num_channel;
    const int *packetdef = Packet::def[type][count];
    for (int i = 0; i < Flit::MAX_FLIT_WORDS; i++) {
        out_flit->data_valid[i] = (packetdef[i] != 0);
        if (packetdef[i] == FT_HEADER) {
            uint dest;
            if (lpmap.find(cur_cmd.remote_id) == lpmap.end()) {
                dest = cur_cmd.remote_id;
            } else {
                dest = lpmap[cur_cmd.remote_id];
            }
            out_flit->data[i]  = (type << 16) | dest;
            out_flit->outclass = out_flit->outch;
            if (lookahead_rc)
                out_flit->nextoutdir = RouteComp::getdir(my_id, dest);
        } else if (packetdef[i] == FT_ADDR_L) {
            out_flit->data[i] = cur_cmd.remote_addr & 0xffffffff;
        } else if (packetdef[i] == FT_ADDR_H) {
            out_flit->data[i] = cur_cmd.remote_addr >> 32;
        } else if (packetdef[i] == FT_STRIDE) {
            out_flit->data[i] = cur_cmd.remote_stride;
        } else if (packetdef[i] == FT_DATA) {
            mem->read(next_cur_cmd.local_addr, 4, &out_flit->data[i]);
            //std::cout << "S a:" << std::hex << next_cur_cmd.local_addr 
            //          << " d:" << out_flit->data[i] << std::dec << std::endl;
            next_cur_cmd.local_addr  += cur_cmd.local_stride;
            next_cur_cmd.remote_addr += cur_cmd.remote_stride;
            next_cur_cmd.transfered += 4;
        } else if (packetdef[i] == FT_IMM) {
            out_flit->data[i] = cur_cmd.immediate;
            next_cur_cmd.transfered += 4;
        }

        if (next_cur_cmd.transfered == cur_cmd.size) { // the command ends
            out_flit->tail = true;
            next_cur_cmd.type = DMA_NA;
            if (cur_cmd.type == DMA_PUT || cur_cmd.type == DMA_PUT64) {
                next_put_cmd.type = DMA_NA;
                mem->write(SC_MMAP_DMA_PUT_CMD, 4, DMA_NA);
            } else if (cur_cmd.type == DMA_GET) {
                deq_get_cmd = true;
            } else if (cur_cmd.type == DMA_ACK) {
                deq_ack_cmd = true;
            }
            break;
        }
    }
    next_count = (out_flit->tail) ? 0 : count + 1;
}

/**********************************************************************/
void InccSend::drive()
{
    // memory-mapped register -> command(s)
    uint request;
    mem->read(SC_MMAP_DMA_PUT_CMD, 4, &request);
    if (request != DMA_NA && ! put_cmd.valid())
        addputcommand();
    mem->read(SC_MMAP_DMA_GET_CMD, 4, &request);
    if (request != DMA_NA) {
        addgetcommand();
        mem->write(SC_MMAP_DMA_GET_CMD, 4, DMA_NA);
    }
    
    // command -> packet(s)
    out_flit->clear();
    out_flit->id    = num_sent_flit;
    out_flit->timer = cycle;
    if (! cur_cmd.valid()) {
        getnextcommand();
    } else {
        uint ch = cur_cmd.channel % num_channel;
        bool send_ok = (count == 0) ? out_flow->accepthead(ch) :
            out_flow->acceptbody(ch);
        if (send_ok)
            generateflit();
    }
}

/**********************************************************************/
void InccSend::update()
{
    type  = next_type;
    count = next_count;
    cur_cmd = next_cur_cmd;
    put_cmd = next_put_cmd;
    if (new_get_cmd.valid()) {
        if (get_cmd.size() != (uint) MAX_ACCEPT_GET)
            get_cmd.push_back(new_get_cmd);
        new_get_cmd.type = DMA_NA;
    }
    if (deq_get_cmd) {
        if (! get_cmd.empty())
            get_cmd.erase(get_cmd.begin());
        deq_get_cmd = false;
    }
    
    if (new_ack_cmd.valid()) {
        ack_cmd.push_back(new_ack_cmd);
        new_ack_cmd.type = DMA_NA;
    }
    if (deq_ack_cmd) {
        if (! ack_cmd.empty())
            ack_cmd.erase(ack_cmd.begin());
        deq_ack_cmd = false;
    }
    cycle++;
    if (out_flit->valid)
        num_sent_flit++;
}

/**********************************************************************/
bool InccSend::isbusy()
{
    uint request;
    if (cur_cmd.valid() || !get_cmd.empty() || !ack_cmd.empty())
        return true;
    mem->read(SC_MMAP_DMA_PUT_CMD, 4, &request);
    if (request != DMA_NA)
        return true;
    mem->read(SC_MMAP_DMA_GET_CMD, 4, &request);
    if (request != DMA_NA)
        return true;
    return false;
}

/* InccRecv                                                           */
/**********************************************************************/
void InccRecv::init(SystemConfig &sc, MainMemory *mem)
{
    this->mem = mem;
    num_channel = sc.num_channel;

    type.resize(num_channel, 0);
    next_type.resize(num_channel, 0);
    count.resize(num_channel, 0);
    next_count.resize(num_channel, 0);
    cmd.resize(num_channel);
    next_cmd.resize(num_channel);
}

/**********************************************************************/
void InccRecv::connect(Incc *incc)
{
    this->in_flow = incc->in_flow[0];
    this->in_flit = incc->in_flit[0];
}

/**********************************************************************/
void InccRecv::readflit()
{
    int ch = flitbuf.inch;
    const int *packetdef = Packet::def[type[ch]][count[ch]];
    for (int i = 0; i < Flit::MAX_FLIT_WORDS; i++) {
        if (! flitbuf.data_valid[i])
            continue;
        if (packetdef[i] == FT_HEADER) {
            next_type[ch] = (flitbuf.data[i] >> 16) % Packet::NUM_TYPE;
            packetdef = Packet::def[next_type[ch]][count[ch]];
            next_cmd[ch].local_addr   = 0;
            next_cmd[ch].local_stride = 4;
        } else if (packetdef[i] == FT_ADDR_L) {
            next_cmd[ch].local_addr   = flitbuf.data[i];
        } else if (packetdef[i] == FT_ADDR_H) {
            next_cmd[ch].local_addr  |= (uint064_t) flitbuf.data[i] << 32;
        } else if (packetdef[i] == FT_STRIDE) {
            next_cmd[ch].local_stride = flitbuf.data[i];
        } else if (packetdef[i] == FT_DATA || packetdef[i] == FT_IMM) {
            mem->write(next_cmd[ch].local_addr, 4, flitbuf.data[i]);
            //std::cout << "R a:" << std::hex << next_cmd[ch].local_addr
            //          << " d:" << flitbuf.data[i] << std::dec << std::endl;
            next_cmd[ch].local_addr += cmd[ch].local_stride;
        } else if (packetdef[i] == FT_END) {
            std::cout << "!! Router: wrong flit ";
            flitbuf.print();
        }
    }
    next_count[ch] = (flitbuf.tail) ? 0 : count[ch] + 1;
}

/**********************************************************************/
void InccRecv::drive()
{
    if (flitbuf.valid)
        readflit();
}

/**********************************************************************/
void InccRecv::update()
{
    type  = next_type;
    count = next_count;
    cmd   = next_cmd;
    flitbuf = *in_flit;
    flitbuf.indir   = flitbuf.outdir;
    flitbuf.inch    = flitbuf.outch;
    flitbuf.inclass = flitbuf.outclass;
    flitbuf.outdir  = flitbuf.nextoutdir;
}

/**********************************************************************/
bool InccRecv::isbusy()
{
    if (flitbuf.valid)
        return true;
    for (uint i = 0; i < num_channel; i++)
        if (count[i] != 0)
            return true;
    return false;
}

/**********************************************************************/
