/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "router.h"
#include "routecomp.h"
#include "../common/id.h"
#include <stdlib.h>

#define FOREACH_CHANNEL(d, c) for(int d = 0; d < NUM_DIR; d++) \
                              for(int c = 0; c < num_channel; c++)

/**********************************************************************/
void Router::init(uint pid, SystemConfig &sc)
{
    Connectable::init(NUM_DIR, sc);
    my_x         = IDX(pid);
    my_y         = IDY(pid);
    my_id        = pid;
    num_channel  = sc.num_channel;
    num_channel  = (num_channel < MAX_VC) ? num_channel : MAX_VC;
    num_inbuf    = sc.num_inbuf;
    lookahead_rc = sc.lookahead_rc;
    credit_fc    = sc.credit_fc;
    skip_lt      = sc.skip_lt;
    lastfree     = false;
    va_allocator.init(NUM_DIR * num_channel, sc.fix_priority);
    va_request.assign(NUM_DIR * num_channel,
                      std::vector<int>(NUM_DIR * num_channel, 0));
    va_grant.assign  (NUM_DIR * num_channel,
                      std::vector<int>(NUM_DIR * num_channel, 0));
    sa_allocator.init(NUM_DIR, sc.fix_priority);
    sa_request.assign(NUM_DIR, std::vector<int>(NUM_DIR, 0));
    sa_grant.assign  (NUM_DIR, std::vector<int>(NUM_DIR, 0));
    for (int i = 0; i < NUM_DIR; i++) {
        mbuf[i].clear();
        next_mbuf[i].clear();
        obuf[i].clear();
        next_obuf[i].clear();
        sa_ch_arbiter[i].init  (num_channel, sc.fix_priority);
        sa_ch_request[i].resize(num_channel, 0);
        sa_ch_grant[i].resize  (num_channel, 0);
    }
    if (va_channel)
        delete va_channel;
    if (sc.fix_vc)
        va_channel = new VCRequestFix(num_channel);
    else
        va_channel = new VCRequestAll(num_channel);

    for (int i = 0; i < NUM_DIR; i++)
        for (int j = 0; j < NUM_DIR; j++)
            xbar_ctrl[i][j] = next_xbar_ctrl[i][j] = false;

    FOREACH_CHANNEL(i, j) {
        ibuf[i][j].clear();
        outdir[i][j]   = next_outdir[i][j] = 0;
        outch[i][j]    = next_outch[i][j] = 0;
        ic_state[i][j] = IC_HEAD;
        oc_state[i][j] = OC_FREE;
        va_success_in[i][j] = va_success_out[i][j] = false;
    }
}

/**********************************************************************/
void Router::driveRC()
{
    FOREACH_CHANNEL(d, c) {
        if (ic_state[d][c] == IC_RC) {
            Flit &head = ibuf[d][c].front();
            uint did   = head.data[0] & 0xffff;
            int  cdir  = (lookahead_rc) ? head.outdir : DIR_C;
            int  ccls  = head.inclass;
            int  ndir  = RouteComp::getdir(my_id, did, cdir, ccls);
            if (lookahead_rc) {
                // outdir[d][c] was already computed in updateLT
                head.nextoutdir = ndir;
            } else {
                next_outdir[d][c] = ndir;
                head.outdir = ndir;
            }
        }
    }
}

/**********************************************************************/
void Router::updateRC()
{
    if (! lookahead_rc) {
        FOREACH_CHANNEL(d, c) {
            outdir[d][c] = next_outdir[d][c];
        }
    }
}

/**********************************************************************/
void Router::clearRC() {}

/**********************************************************************/
void Router::driveVA()
{
    int request_dir[NUM_DIR][MAX_VC];
    FOREACH_CHANNEL(d, c) {
        bool en = ((ic_state[d][c] == IC_RC && lookahead_rc) ||
                   (ic_state[d][c] == IC_VA));
        request_dir[d][c] = (en) ? outdir[d][c] : -1;
    }

    FOREACH_CHANNEL(sd, sc) {
        int sp     = sd * num_channel + sc;
        if (request_dir[sd][sc] != -1) {
            int sclass = ibuf[sd][sc].front().inclass;
            FOREACH_CHANNEL(dd, dc) {
                int dp = dd * num_channel + dc;
                va_request[sp][dp] = ((request_dir[sd][sc] == dd) &&
                                      (va_channel->ok(sclass, dc)) &&
                                      (oc_state[dd][dc] == OC_FREE));
            }
        }
    }

    va_allocator.allocate(va_request, va_grant);

    FOREACH_CHANNEL(sd, sc) {
        int sp = sd * num_channel + sc;
        FOREACH_CHANNEL(dd, dc) {
            int dp = dd * num_channel + dc;
            if (va_grant[sp][dp]) {
                va_success_in [sd][sc] = true;
                va_success_out[dd][dc] = true;
                next_outch[sd][sc] = dc;
            }
        }
    }
}

/**********************************************************************/
void Router::updateVA()
{
    FOREACH_CHANNEL(d, c) {
        outch[d][c] = next_outch[d][c];
    }
}

/**********************************************************************/
void Router::clearVA() {
    for (int i = 0; i < NUM_DIR * num_channel; i++)
        for (int j = 0; j < NUM_DIR * num_channel; j++)
            va_request[i][j] = va_grant[i][j] = 0;
    FOREACH_CHANNEL(d, c) {
        va_success_in[d][c] = va_success_out[d][c] = false;
    }
}

/**********************************************************************/
void Router::driveSA()
{
    // SA input generator
    bool sa_enable[NUM_DIR][MAX_VC];
    FOREACH_CHANNEL(sd, sc) {
        int dd = outdir[sd][sc];
        int dc = outch [sd][sc];
        sa_enable[sd][sc] = (ic_state[sd][sc] == IC_ACTIVE && 
                             ! ibuf[sd][sc].empty() && 
                             acceptbody(dd, dc));
        sa_request[sd][dd] = (sa_enable[sd][sc]) ? 1 : sa_request[sd][dd];
    }

    sa_allocator.allocate(sa_request, sa_grant);

    for (int sd = 0; sd < NUM_DIR; sd++)
        for (int dd = 0; dd < NUM_DIR; dd++)
           next_xbar_ctrl[sd][dd] = (sa_grant[sd][dd] != 0);

    // SA output generator
    FOREACH_CHANNEL(d, c) {
        sa_ch_request[d][c] = sa_enable[d][c] && sa_grant[d][outdir[d][c]];
    }

    for (int d = 0; d < NUM_DIR; d++) {
        sa_ch_arbiter[d].arbitrate(sa_ch_request[d], sa_ch_grant[d]);
        for (int c = 0; c < num_channel; c++)
            if (sa_ch_grant[d][c]) {
                next_mbuf[d] = ibuf[d][c].front();
                next_mbuf[d].outch    = outch[d][c];
                next_mbuf[d].outdir   = outdir[d][c];
            }
    }
}

/**********************************************************************/
void Router::updateSA()
{
    for (int sd = 0; sd < NUM_DIR; sd++) {
        for (int dd = 0; dd < NUM_DIR; dd++)
            xbar_ctrl[sd][dd] = next_xbar_ctrl[sd][dd];
        mbuf[sd] = next_mbuf[sd];
        int dd = mbuf[sd].outdir;
        int dc = mbuf[sd].outch;
        for (int sc = 0; sc < num_channel; sc++)
            if (sa_ch_grant[sd][sc]) {
                ibuf[sd][sc].erase(ibuf[sd][sc].begin());
                if (credit_fc) {
                    inccredit(sd, sc);
                    deccredit(dd, dc);
                } else if ((int) ibuf[sd][sc].size() < num_inbuf - 2) {
                    turnon(sd, sc);
                }
            }
    }
}

/**********************************************************************/
void Router::clearSA() 
{
    for (int sd = 0; sd < NUM_DIR; sd++)
        for (int dd = 0; dd < NUM_DIR; dd++) {
            sa_request[sd][dd] = sa_grant[sd][dd] = 0;
            next_xbar_ctrl[sd][dd] = 0;
        }
    FOREACH_CHANNEL(d, c) {
        sa_ch_request[d][c] = sa_ch_grant[d][c] = 0;
    }
    for (int d = 0; d < NUM_DIR; d++)
        next_mbuf[d].clear();

}

/**********************************************************************/
void Router::driveST()
{
    for (int dd = 0; dd < NUM_DIR; dd++) {
        for (int sd = 0; sd < NUM_DIR; sd++)
            if (xbar_ctrl[sd][dd])
                next_obuf[dd] = mbuf[sd];
        if (skip_lt)
            *out_flit[dd] = next_obuf[dd];
    }
}

/**********************************************************************/
void Router::updateST()
{
    for (int dd = 0; dd < NUM_DIR; dd++)
        obuf[dd] = next_obuf[dd];
}

/**********************************************************************/
void Router::clearST()
{
    for (int dd = 0; dd < NUM_DIR; dd++)
        next_obuf[dd].clear();
}

/**********************************************************************/
void Router::driveLT()
{
    if (! skip_lt)
        for (int dd = 0; dd < NUM_DIR; dd++)
            *out_flit[dd] = obuf[dd];
}

/**********************************************************************/
void Router::updateLT()
{
    for (int d = 0; d < NUM_DIR; d++) {
        if (! in_flit[d] || ! in_flit[d]->valid)
            continue;
        pushflit(d, in_flit[d]->outch, *in_flit[d]);
    }
}

/**********************************************************************/
void Router::clearLT() {}

/**********************************************************************/
void Router::drive()
{
    // if there are no flits, only disable output signals
    if (lastfree) {
        for (int d = 0; d < NUM_DIR; d++)
            out_flit[d]->clear();
        return;
    }

    // if there are flits, drive each stage
    driveRC();
    driveVA();
    driveSA();
    driveST();
    driveLT();
}

/**********************************************************************/
void Router::update()
{
    // if there are no flits, only see if a flit comes
    if (lastfree) {
        bool noinput = true;
        for (int d = 0; d < NUM_DIR; d++) {
            if (in_flit[d] && in_flit[d]->valid)
                noinput = false;
        }
        if (noinput)
            return;
    }
        
    // update each stage (RC to ST can be skipped if there are no flits)
    updateRC();
    updateVA();
    updateSA();
    updateST();
    updateLT();

    // update states
    FOREACH_CHANNEL(d, c) {
        if (ic_state[d][c] == IC_HEAD)
            ic_state[d][c] = (! ibuf[d][c].empty()) ? IC_RC : IC_HEAD;
        else if (ic_state[d][c] == IC_RC || ic_state[d][c] == IC_VA)
            ic_state[d][c] = (va_success_in[d][c]) ? IC_ACTIVE : IC_VA;
        else if (ic_state[d][c] == IC_ACTIVE)
            ic_state[d][c] = (mbuf[d].inch == c && mbuf[d].tail) ?
                IC_TAIL : IC_ACTIVE;
        else /* if (ic_state[d][c] == IC_TAIL) */
            ic_state[d][c] = IC_HEAD;

        if (oc_state[d][c] == OC_FREE)
            oc_state[d][c] = (va_success_out[d][c]) ? OC_LOCK : OC_FREE;
        else if (oc_state[d][c] == OC_LOCK)
            oc_state[d][c] = (obuf[d].outch == c && obuf[d].tail) ?
                OC_TAIL : OC_LOCK;
        else /* if (oc_state[d][c] == OC_TAIL) */
            oc_state[d][c] = (accepthead(d, c)) ? OC_FREE : OC_TAIL;
    }

    // clear wire signals
    clearRC();
    clearVA();
    clearSA();
    clearST();
    clearLT();
    lastfree = ! isbusy();
}

/**********************************************************************/
void Router::pushflit(int dir, int ch, Flit flit)
{
    flit.indir    = flit.outdir;
    flit.inch     = flit.outch;
    flit.inclass  = flit.outclass;
    flit.outdir   = flit.nextoutdir;
    flit.outch    = num_channel;
    flit.nextoutdir = NUM_DIR;
    if (lookahead_rc && flit.head)
        outdir[dir][ch] = flit.outdir;
    ibuf[dir][ch].push_back(flit);
    if (! credit_fc && (int) ibuf[dir][ch].size() >= num_inbuf - 2)
        turnoff(dir, ch);
}

/**********************************************************************/
bool Router::isbusy()
{
    for (int d = 0; d < NUM_DIR; d++) {
        if (mbuf[d].valid || obuf[d].valid)
            return true;
        for (int c = 0; c < num_channel; c++)
            if (! ibuf[d][c].empty())
                return true;
    }
    return false;
}

/**********************************************************************/
void Router::print()
{
    std::cout << "X:" << my_x << " Y:" << my_y << std::endl;
    for (int d = 0; d < NUM_DIR; d++) {
        for (int c = 0; c < num_channel; c++) {
            std::cout << 'd' << d << 'c' << c;
            std::cout << " I:" << ic_state[d][c] << " [";
            for (int b = num_inbuf - 1; b >= 0; b--) {
                if ((int) ibuf[d][c].size() > b)
                    std::cout << ((ibuf[d][c][b].head) ? 'h' :
                                  (ibuf[d][c][b].tail) ? 't' : 'b');
                else 
                    std::cout << '-';
            }
            if (c == num_channel - 1) {
                std::cout <<   "] [" << ((! mbuf[d].valid) ? '-' :
                                         (mbuf[d].head) ? 'h' :
                                         (mbuf[d].tail) ? 't' : 'b');
                std::cout << "] X [" << ((! obuf[d].valid) ? '-' :
                                         (obuf[d].head) ? 'h' :
                                         (obuf[d].tail) ? 't' : 'b');
                std::cout << ']';
            } else {
                std::cout << "]          ";
            }
            std::cout << " O:" << oc_state[d][c] << std::endl;
        }
    }
}

/**********************************************************************/
