/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#pragma once

#include "connectable.h"
#include "packet.h"
#include "allocator.h"
#include "vcrequest.h"
#include "../common/config.h"
#include <vector>

/* Virtual Channel Router for M-Core **********************************/
/* Pipeline Stage     : 3, 4 or 5 stages                              */
/*                      3 or 4 stages: NRC+VA / SA / ST (/ LT)        */
/*                      4 or 5 stages: RC / VA / SA / ST (/ LT)       */
/* Routing            : X-Y Dimension Order, Warmhole                 */
/* Flow Control       : Credit-base (credit_fc) or Xon/Xoff           */
/* VC & SW Allocator  : iSlip or Fixed Priority (fix_priority)        */
/* FIFO Depth         : Variable (num_inbuf)                          */
/* # Virtual Channels : Variable (num_channel, up to MAX_VC)          */
/* Data Width         : 32 (use packet32) or 128 (use packet128)      */
/**********************************************************************/
class Router : public Connectable {
 private:
    typedef enum {
        IC_HEAD   = 0,
        IC_RC     = 1,
        IC_VA     = 2,
        IC_ACTIVE = 3,
        IC_TAIL   = 4,
    } ics_t;
    typedef enum {
        OC_FREE   = 0,
        OC_LOCK   = 1,
        OC_TAIL   = 2,
    } ocs_t;

    std::vector<Flit> ibuf[NUM_DIR][MAX_VC]; // input buffer
    Flit mbuf[NUM_DIR], next_mbuf[NUM_DIR];  // middle (SA/ST) buffer
    Flit obuf[NUM_DIR], next_obuf[NUM_DIR];  // output (ST/LT) buffer
    Allocator va_allocator, sa_allocator;    // VA, SA
    Arbiter   sa_ch_arbiter[NUM_DIR];        // SA output channel
    std::vector<std::vector<int> > va_request, va_grant;
    std::vector<std::vector<int> > sa_request, sa_grant;
    std::vector<int> sa_ch_request[NUM_DIR], sa_ch_grant[NUM_DIR];
    VCRequest *va_channel;             // Packet Class -> Request Channel

    bool  xbar_ctrl[NUM_DIR][NUM_DIR], next_xbar_ctrl[NUM_DIR][NUM_DIR];
    int   outdir   [NUM_DIR][MAX_VC],  next_outdir   [NUM_DIR][MAX_VC];
    int   outch    [NUM_DIR][MAX_VC],  next_outch    [NUM_DIR][MAX_VC];
    ics_t ic_state [NUM_DIR][MAX_VC];
    ocs_t oc_state [NUM_DIR][MAX_VC];
    bool  va_success_in[NUM_DIR][MAX_VC], va_success_out[NUM_DIR][MAX_VC];

    uint my_x, my_y, my_id;
    int  num_channel, num_inbuf;
    bool lookahead_rc, credit_fc, skip_lt;
    bool lastfree; // keep if the router was free last cycle (for speedup)

    void driveRC();
    void driveVA();
    void driveSA();
    void driveST();
    void driveLT();
    void updateRC();
    void updateVA();
    void updateSA();
    void updateST();
    void updateLT();
    void clearRC();
    void clearVA();
    void clearSA();
    void clearST();
    void clearLT();

 public:
    Router() { va_channel = NULL; }
    virtual ~Router() { if (va_channel) delete va_channel; }
    void init (uint pid, SystemConfig &sc);
    void drive();
    void update();
    void pushflit(int dir, int ch, Flit flit);
    bool isbusy();
    void print();
};

/**********************************************************************/
