/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "nocenv.h"
#include "common/config.h"
#include "common/rand.h"
#include "version.h"
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <sstream>

const double Env::RATE_INTERVAL = 0.05;
const double Env::END_RATE = 1.00;

/**********************************************************************/
int main (int argc, char **argv)
{
    Env env;
    std::cout << "## " << MC_NOC << ' ' << MC_VER << std::endl;
    return env.sim_main(argc, argv);
}

/**********************************************************************/
Env::Env()
{
    packet_length     = PACKET_LENGTH_DEF;
    pattern_str       = "RANDOM";
    rate_interval     = RATE_INTERVAL;
    rate_end          = END_RATE;
    use_smart         = false;
    smart_pattern_str = "INTERLEAVE";
    use_normalized    = true;
    end_saturated     = true;
    end_once          = false;
}

/**********************************************************************/
void Env::usage()
{   
    static const char *usagemessage = "\
usage: ./SimMc-NoC [options]\n\
  --x-rank num , -x[num]: number of columns \n\
  --y-rank num , -y[num]: number of rows \n\
  --channel num, -c[num]: number of virtual channels \n\
  --buffer num , -b[num]: length of input buffers (in flit) \n\
  --packet num , -p[num]: length of packets (in flit) \n\
  --pattern string      : traffic pattern\n\
      available patterns are RANDOM, COMPLEMENT, and TORNADO. \n\
  --(no-)lookahead-rc,  : enable/disable lookahead route computation \n\
  --(no-)fix-vc         : enable/disable fixed VC assignment \n\
  --(no-)fix-priority   : enable/disable fixed allocation priority \n\
  --(no-)credit-fc      : enable/disable credit-based flow control \n\
  --(no-)link-traversal : enable/disable LT stage \n\
  --(no-)smart          : enable/disable SmartCore System\n\
                          -c and --(no-)fix_vc options will be ignored\n\
  --smart-pattern string: node allocation pattern for SmartCore\n\
      available patterns are INTERLEAVE, CHECK and BLOCK\n\
  --rate num   , -r[num]: injection rate (interval) \n\
  --rate-end num        : maximum injection rate\n\
  --(no-)normalize      : use normalized injection rate\n\
  --(no-)end-saturated  : finish simulation if throughput is saturated\n\
  --(no-)sim-once       : simulate only once\n\
  --help                : show this message\n\
  --version             : show version \n\
\n\
default: -x8 -y8 -c2 -b4 -p5 --pattern RANDOM --lookahead-rc \n\
         --no-fix-vc --no-fix-priority --credit-fc --link-traversal \n\
         --no-smart -r0.05 --rate-end 1.00 --normalize --end-saturated\n\
         --no-sim-once \n\n\
";
    std::cout << usagemessage;
}

/**********************************************************************/
bool Env::wrong_option (char *buf)
{
    std::cerr << "!! wrong option: " << buf << std::endl;
    usage();
    return false;
}

/**********************************************************************/
double Env::readargnum (char *buf)
{
    if (buf == NULL)
        return -1;
    std::istringstream iss(buf);
    double d = 0;
    iss >> d;
    return d;
}

/**********************************************************************/
bool Env::readargs (int argc, char **argv)
{
    int index = 1;
    std::cout << "## options:";
    while (index < argc) {
        int headindex = index;
        std::cout << ' ' << argv[index];
        if (argv[index][0] == '-') {
            if (argv[index][1] == 'x')
                sc.x_rank      = (int) readargnum(&argv[index][2]);
            else if (argv[index][1] == 'y')
                sc.y_rank      = (int) readargnum(&argv[index][2]);
            else if (argv[index][1] == 'c')
                sc.num_channel = (int) readargnum(&argv[index][2]);
            else if (argv[index][1] == 'b')
                sc.num_inbuf   = (int) readargnum(&argv[index][2]);
            else if (argv[index][1] == 'p')
                packet_length  = (int) readargnum(&argv[index][2]);
            else if (argv[index][1] == 'r')
                rate_interval  =       readargnum(&argv[index][2]);
            else if (argv[index][1] == '-') {
                std::string argname(&argv[index][2]);
                bool value = true;
                if (argname.find("no-", 0) == 0) {
                    argname.erase(0, 3);
                    value = false;
                }
                if      (argname == "x-rank")
                    sc.x_rank         = (int) readargnum(argv[++index]);
                else if (argname == "y-rank")
                    sc.y_rank         = (int) readargnum(argv[++index]);
                else if (argname == "channel")
                    sc.num_channel    = (int) readargnum(argv[++index]);
                else if (argname == "buffer")
                    sc.num_inbuf      = (int) readargnum(argv[++index]);
                else if (argname == "packet")
                    packet_length     = (int) readargnum(argv[++index]);
                else if (argname == "rate")
                    rate_interval     =       readargnum(argv[++index]);
                else if (argname == "rate-end")
                    rate_end          =       readargnum(argv[++index]);
                else if (argname == "pattern")
                    pattern_str       = (argv[++index]) ? argv[index] : "NULL";
                else if (argname == "lookahead-rc")
                    sc.lookahead_rc   = value;
                else if (argname == "fix-vc")
                    sc.fix_vc         = value;
                else if (argname == "fix-priority")
                    sc.fix_priority   = value;
                else if (argname == "credit-fc")
                    sc.credit_fc      = value;
                else if (argname == "link-traversal")
                    sc.skip_lt        = !value;
                else if (argname == "smart")
                    use_smart         = value;
                else if (argname == "smart-pattern")
                    smart_pattern_str = (argv[++index]) ? argv[index] : "NULL";
                else if (argname == "normalize")
                    use_normalized    = value;
                else if (argname == "end-saturated")
                    end_saturated     = value;
                else if (argname == "sim-once")
                    end_once          = value;
                else if (argname == "help") {
                    std::cout << std::endl;
                    usage();
                    exit(0);
                } else if (argname == "version") {
                    std::cout << std::endl;
                    exit(0);
                } else
                    return wrong_option(argv[index]);
            } else {
                return wrong_option(argv[index]);
            }
        } else {
            return wrong_option(argv[index]);
        }
        if (headindex != index)
            std::cout << ' ' << argv[index];
        index++;
    }
    std::cout << std::endl;
    return true;
}

/**********************************************************************/
double Env::network_capacity(int x, int y, bool smart)
{
    int xx = (smart) ? x / 2 : x;
    int bisection_bw = (x < y) ? x : y;
    return (double) bisection_bw / (xx * y / 2) * 2;
}

/**********************************************************************/
double Env::sim (double rate)
{
    double capacity   = ((! use_normalized) ? 1 :
                         network_capacity(sc.x_rank, sc.y_rank, use_smart));
    ullint warm_cycle = WARM_UP_CYCLE_DEF;
    ullint main_cycle = SIM_CYCLE_DEF;

    // initialize
    int num_node = sc.x_rank * sc.y_rank;
    core.resize  (num_node);
    router.resize(num_node);
    if (use_smart)
        smart.resize(num_node);
    for (int i = 0; i < num_node; i++) {
        int x = i % sc.x_rank + 1;
        int y = i / sc.x_rank + 1;
        core[i].init  (ID(x, y), sc);
        router[i].init(ID(x, y), sc);
        if (!use_smart) {
            core[i].connect(0, router[i], DIR_C);
            router[i].connect(DIR_C, core[i], 0);
        } else {
            smart[i].init(ID(x, y), sc, &router[i]);
            core[i].connect(0, smart[i], SmartUnit::SIDE_INCC);
            router[i].connect(DIR_C, smart[i], SmartUnit::SIDE_RT);
            smart[i].connect(SmartUnit::SIDE_INCC, core[i], 0);
            smart[i].connect(SmartUnit::SIDE_RT, router[i], DIR_C);
        }
        core[i].warm_cycle = warm_cycle;
        core[i].main_cycle = warm_cycle + main_cycle;
    }
    for (int i = 0; i < num_node; i++) {
        int x = i % sc.x_rank + 1;
        int y = i / sc.x_rank + 1;
        if (x != sc.x_rank) {
            router[i].connect(DIR_E, router[i + 1], DIR_W);
            router[i + 1].connect(DIR_W, router[i], DIR_E);
        }
        if (y != sc.y_rank) {
            router[i].connect(DIR_S, router[i + sc.x_rank], DIR_N);
            router[i + sc.x_rank].connect(DIR_N, router[i], DIR_S);            
        }
    }

    // simulation
    ullint total_send, total_recv, total_recv_main, total_latency, cycle;
    total_send = total_recv = cycle = 0;
    while (cycle < warm_cycle + main_cycle || total_send != total_recv) {
        total_send = total_recv = total_recv_main = total_latency = 0;
        for (int i = 0; i < num_node; i++) {
            core[i].drive();
            router[i].drive();
            if (use_smart)
                smart[i].drive();
        }
        for (int i = 0; i < num_node; i++) {
            int x = i % sc.x_rank + 1;
            int y = i / sc.x_rank + 1;
            core[i].update();
            router[i].update();
            if (use_smart)
                smart[i].update();
            if (!use_smart || sc.ismaster(ID(x, y))) {
                total_send      += core[i].num_send;
                total_recv      += core[i].num_recv;
                total_recv_main += core[i].num_recv_main;
                total_latency   += core[i].sum_latency;
            }
        }
        for (int i = 0; i < sc.num_core; i++) {
            if (RandGen::generate() < rate / packet_length * capacity) {
                int w = (use_smart) ? sc.x_rank / 2 : sc.x_rank;
                int h = sc.y_rank;
                int x = i % w + 1;
                int y = i / w + 1;
                uint srcid, destid;
                srcid = destid = ID(x, y);
                if (pattern == RANDOM) {
                    while (destid == ID(x, y))
                        destid = ID(RandGen::generate(w) + 1,
                                    RandGen::generate(h) + 1);
                } else if (pattern == COMPLEMENT) {
                    destid = ID(w - x + 1, h - y + 1);
                } else if (pattern == TORNADO) {
                    destid = ID((x + w / 2 - 1) % w + 1,
                                (y + h / 2 - 1) % h + 1);
                }
                if (! use_smart) {
                    int ch = (IDX(destid) + IDY(destid)) % sc.num_channel;
                    core[i].addpacket(destid, packet_length, ch);
                } else {
                    uint srcid_m = sc.getpid(0, srcid);
                    uint srcid_s = sc.getslavepid(srcid_m);
                    int i_m = IDX(srcid_m) - 1 + (IDY(srcid_m) - 1) * w * 2;
                    int i_s = IDX(srcid_s) - 1 + (IDY(srcid_s) - 1) * w * 2;
                    destid = sc.getpid(0, destid);
                    core[i_m].addpacket(destid, packet_length, 0);
                    core[i_s].addpacket(destid, packet_length, 0);
                }
            }
        }
        cycle++;
    }

    // finalize
    double avg_latency, throughput;
    avg_latency = (double) total_latency / total_recv;
    throughput  = ((double) total_recv_main / sc.num_core 
                   * packet_length / main_cycle / capacity);
    std::cout << std::fixed << std::setprecision(3) << rate << '\t';
    std::cout << std::fixed << std::setprecision(3) << throughput << '\t';
    std::cout << std::fixed << std::setprecision(2) << avg_latency;
    std::cout << std::endl;
    core.resize(0);
    router.resize(0);
    smart.resize(0);
    return throughput;
}

/**********************************************************************/
int Env::sim_main (int argc, char **argv)
{
    // default settings
    sc.x_rank       = SIZE_X_DEF;
    sc.y_rank       = SIZE_Y_DEF;
    sc.num_channel  = NUM_CHANNEL_DEF;
    sc.num_inbuf    = BUFFER_LENGTH_DEF;
    sc.lookahead_rc = true;
    sc.fix_vc       = false;
    sc.fix_priority = false;
    sc.credit_fc    = true;
    sc.skip_lt      = false;

    // read & check arguments
    if (! readargs(argc, argv))
        return 1;

    if (sc.x_rank <= 0 || sc.y_rank <= 0 || sc.x_rank * sc.y_rank <= 1) {
        std::cerr << "!! invalid NoC size" << std::endl;
        usage();
        return 1;
    }
    if (sc.num_channel <= 0 || sc.num_inbuf <= 0 || packet_length <= 0) {
        std::cerr << "!! invalid NoC settings" << std::endl;
        usage();
        return 1;
    }

    if      (pattern_str == "RANDOM")
        pattern = RANDOM;
    else if (pattern_str == "COMPLEMENT")
        pattern = COMPLEMENT;
    else if (pattern_str == "TORNADO")
        pattern = TORNADO;
    else {
        std::cerr << "!! invalid pattern: " << pattern_str << std::endl;
        usage();
        return 1;
    }
    if (use_smart) {
        sc.num_channel = 3;
        sc.fix_vc      = true;
    }

    sc.setdefaultconfig(sc.x_rank, sc.y_rank, false, false,
                        use_smart, smart_pattern_str);
    sc.max_packet_length = packet_length;
    std::cout << "offer\tt_put\tlatency" << std::endl;
    double rate, result;
    for (rate = rate_interval; rate <= rate_end; rate += rate_interval) {
        result = sim(rate);
        if (end_once || (end_saturated && result < rate - 0.05))
            break;
    }

    return 0;
}

/**********************************************************************/
