/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "smartunit.h"
#include "../common/id.h"

/**********************************************************************/
void SmartUnit::init(uint pid, SystemConfig &sc, Router *rt)
{
    Connectable::init(2, sc);
    this->rt     = rt;
    master_mode  = sc.ismaster(pid);
    slave_mode   = sc.isslave(pid);
    my_id        = pid;
    my_x         = IDX(pid);
    my_y         = IDY(pid);
    if (master_mode)
        pair_id  = sc.getslavepid(pid);
    else if (slave_mode)
        pair_id  = sc.getmasterpid(pid);
    else
        pair_id  = pid;
    pair_x       = IDX(pair_id);
    pair_y       = IDY(pair_id);
    num_channel  = sc.num_channel;
    num_inbuf    = SU_BUFFER_LENGTH_DEF;
    num_inbuf_rt = sc.num_inbuf;
    lookahead_rc = sc.lookahead_rc;
    credit_fc    = sc.credit_fc;

    if (num_inbuf < num_inbuf_rt)
        num_inbuf = num_inbuf_rt;
    if (num_inbuf < sc.max_packet_length + ((credit_fc) ? 0 : 2))
        num_inbuf = sc.max_packet_length + ((credit_fc) ? 0 : 2);

    flow_m.init(1, num_inbuf);
    flow_s.init(1, num_inbuf);
    vbuf_deq      = false;
    inch          = 0;
    inch_lock     = false;
    timeout_cycle = SU_TIMEOUT_DEF;
    elapsed_cycle = 0;
    total_cycle   = 0;
    detect_error  = false;
}

/**********************************************************************/
int SmartUnit::getcredit_master(int ch, FlowControl &flow)
{
    // calculate the next credit
    int credit_rt = getcredit(SIDE_RT, ch);
    int credit_su = flow.getcredit();
    int newcredit = (credit_rt < credit_su) ? credit_rt : credit_su;
    // do not accept the next head flit until the buffer becomes empty
    if (newcredit == num_inbuf_rt && credit_su != num_inbuf)
        newcredit = num_inbuf_rt - 1;
    return newcredit;
}

/**********************************************************************/
void SmartUnit::drive_master()
{
    // verify buffer
    if (! vbuf_m.empty() && ! vbuf_s.empty()) {
        Flit &fm = vbuf_m.front();
        Flit &fs = vbuf_s.front();
        bool compare = true;
        for (int i = 0; i < Flit::MAX_FLIT_WORDS; i++) {
            uint032_t fmd = fm.data[i];
            uint032_t fsd = fs.data[i];
            fmd = (fm.head && i == 0) ? fmd & 0xffff0000 : fmd;
            fsd = (fs.head && i == 0) ? fsd & 0xffff0000 : fsd;
            if (fm.data_valid[i] && fs.data_valid[i] && fmd != fsd)
                compare = false;
        }
        if (! compare && ! detect_error) {
            std::cout << "!! SC-Master: flit mismatch at cycle "
                      << (total_cycle + 1) << std::endl;
            std::cout << "!! (M) ";
            fm.print();
            std::cout << "!! (S) ";
            fs.print();
            detect_error = true;
        }
        elapsed_cycle = 0;
        vbuf_deq = true;
    } else if (! vbuf_m.empty() || ! vbuf_s.empty()) {
        elapsed_cycle++;
        if (elapsed_cycle >= timeout_cycle && ! detect_error) {
            std::cout << "!! SC-Master: timeout at node " << std::hex 
                      << my_id << ", cycle " << std::dec
                      << (total_cycle + 1) << std::endl;
            detect_error = true;
        }
    }
    // copy flit        
    if (in_flit[SIDE_RT]->valid && in_flit[SIDE_RT]->outch == CH_NORMAL) {
        copy_flit = *in_flit[SIDE_RT];
        if (copy_flit.head)
            copy_flit.data[0] = (pair_id & 0x0000ffff) |
                      (copy_flit.data[0] & 0xffff0000);
        copy_flit.outch = CH_COPY;
        copy_flit.outclass = CH_COPY;
        if (lookahead_rc && copy_flit.head)
            copy_flit.nextoutdir = RouteComp::getdir(my_id, pair_id);
    }

    // Send (INCC -> Router)
    *out_flit[SIDE_RT]   = *in_flit[SIDE_INCC];
    if (out_flit[SIDE_RT]->valid) {
        out_flit[SIDE_RT]->outch = CH_NORMAL;
        out_flit[SIDE_RT]->outclass = CH_NORMAL;
    }

    // Recv (Router -> INCC)
    *out_flit[SIDE_INCC] = *in_flit[SIDE_RT];
    if (out_flit[SIDE_INCC]->valid) {
        if (out_flit[SIDE_INCC]->outch == CH_MERGE) {
            out_flit[SIDE_INCC]->valid = false;
        } else if (out_flit[SIDE_INCC]->outch != CH_NORMAL && ! detect_error) {
            std::cout << "!! SC-Master: wrong flit ";
            out_flit[SIDE_INCC]->print();
            detect_error = true;
        }
    }
}

/**********************************************************************/
void SmartUnit::update_master()
{
    // Enqueue verify buffer
    if (in_flit[SIDE_INCC]->valid) {
        vbuf_m.push_back(*in_flit[SIDE_INCC]);
        if (credit_fc)
            flow_m.deccredit();
    }
    if (in_flit[SIDE_RT]->valid && in_flit[SIDE_RT]->outch == CH_MERGE) {
        vbuf_s.push_back(*in_flit[SIDE_RT]);
        if (credit_fc)
            flow_s.deccredit();
    }
    // copy flit to router's copy channel
    if (copy_flit.valid) {
        rt->pushflit(DIR_C, CH_COPY, copy_flit);
        if (credit_fc)
            rt->in_flow[DIR_C]->deccredit(CH_COPY); // it's ugly...ToDo!
        copy_flit.valid = false;
    }
    // Dequeue verify buffer
    if (vbuf_deq) {
        if (! vbuf_m.empty())
            vbuf_m.erase(vbuf_m.begin());
        if (! vbuf_s.empty())
            vbuf_s.erase(vbuf_s.begin());
        if (credit_fc) {
            flow_m.inccredit();
            flow_s.inccredit();
        }
        vbuf_deq = false;
    }

    // Send (INCC -> Router)
    if (in_flit[SIDE_INCC]->valid) {
        if (in_flit[SIDE_INCC]->head) {
            inch = in_flit[SIDE_INCC]->outch;
            inch_lock = true;
        } else if (in_flit[SIDE_INCC]->tail) {
            inch_lock = false;
        }
        if (credit_fc)
            deccredit(SIDE_RT, CH_NORMAL);
    }
    for (int i = 0; i < num_channel; i++) {
        if (inch_lock && i != inch) {
            turnoff(SIDE_INCC, i);
        } else if (credit_fc) {
            setcredit(SIDE_INCC, i, getcredit_master(CH_NORMAL, flow_m)); 
        } else {
            if (acceptbody(SIDE_RT, CH_NORMAL) &&
                (int) vbuf_m.size() < num_inbuf - 2)
                turnon(SIDE_INCC, i);
            else
                turnoff(SIDE_INCC, i);
        }
    }

    // Recv (Router -> INCC)
    for (int i = 0; i < num_channel; i++)
        if (i == CH_NORMAL) {
            setcredit(SIDE_RT, i, getcredit(SIDE_RT, CH_COPY));
        } else if (i == CH_MERGE) {
            if (credit_fc) {
                setcredit(SIDE_RT, i, getcredit_master(CH_MERGE, flow_s)); 
            } else {
                if ((int) vbuf_s.size() < num_inbuf - 2)
                    turnon(SIDE_RT, i);
                else
                    turnoff(SIDE_RT, i);
            }
        } else {
            turnon(SIDE_RT, i);
        }
}

/**********************************************************************/
void SmartUnit::drive_slave()
{
    // Send (INCC -> Router)
    *out_flit[SIDE_RT]   = *in_flit[SIDE_INCC];
    if (out_flit[SIDE_RT]->valid) {
        if (out_flit[SIDE_RT]->head)
            out_flit[SIDE_RT]->data[0] = (pair_id & 0x0000ffff) |
                      (out_flit[SIDE_RT]->data[0] & 0xffff0000);
        if (out_flit[SIDE_RT]->head && lookahead_rc)
            out_flit[SIDE_RT]->nextoutdir = RouteComp::getdir(my_id, pair_id);
        out_flit[SIDE_RT]->outch = CH_MERGE;
        out_flit[SIDE_RT]->outclass = CH_MERGE;
    }
    // Recv (Router -> INCC)
    *out_flit[SIDE_INCC] = *in_flit[SIDE_RT];
    if (out_flit[SIDE_INCC]->valid) {
        if (out_flit[SIDE_INCC]->outch != CH_COPY && ! detect_error) {
            std::cout << "!! SC-Slave: wrong flit ";
            out_flit[SIDE_INCC]->print();
            detect_error = true;
        }
        out_flit[SIDE_INCC]->outch = CH_NORMAL;
        out_flit[SIDE_INCC]->outclass = CH_NORMAL;
    }
}

/**********************************************************************/
void SmartUnit::update_slave()
{
    // Send (INCC -> Router)
    if (in_flit[SIDE_INCC]->valid) {
        if (in_flit[SIDE_INCC]->head) {
            inch = in_flit[SIDE_INCC]->outch;
            inch_lock = true;
        } else if (in_flit[SIDE_INCC]->tail) {
            inch_lock = false;
        }
        if (credit_fc)
            deccredit(SIDE_RT, CH_MERGE);
    }
    for (int i = 0; i < num_channel; i++) {
        if (inch_lock && i != inch)
            turnoff(SIDE_INCC, i);
        else
            setcredit(SIDE_INCC, i, getcredit(SIDE_RT, CH_MERGE));
    }

    // Recv (Router -> INCC)
    for (int i = 0; i < num_channel; i++)
        turnon(SIDE_RT, i); 
}

/**********************************************************************/
void SmartUnit::drive_normal()
{
    *out_flit[SIDE_RT]   = *in_flit[SIDE_INCC];
    *out_flit[SIDE_INCC] = *in_flit[SIDE_RT];
}

/**********************************************************************/
void SmartUnit::update_normal()
{
    // Send (INCC -> Router)
    for (int i = 0; i < num_channel; i++) {
        if (credit_fc && in_flit[SIDE_INCC]->valid &&
            in_flit[SIDE_INCC]->outch == i)
            deccredit(SIDE_RT, i);
        setcredit(SIDE_INCC, i, getcredit(SIDE_RT, i));
    }

    // Recv (Router -> INCC)
    for (int i = 0; i < num_channel; i++)
        turnon(SIDE_RT, i); 
}

/**********************************************************************/
void SmartUnit::drive()
{   
    if (master_mode) {
        drive_master();
    } else if (slave_mode) {
        drive_slave();
    } else {
        drive_normal();
    }
}

/**********************************************************************/
void SmartUnit::update()
{   
    if (master_mode) {
        update_master();
    } else if (slave_mode) {
        update_slave();
    } else {
        update_normal();
    }
    total_cycle++;
}

/**********************************************************************/
void SmartUnit::print()
{
    std::cout << "X:" << my_x << " Y:" << my_y;
    if (master_mode) {
        std::cout << " Mode:Master" << std::endl;
    } else if (slave_mode) {
        std::cout << " Mode:Slave" << std::endl;
    } else {
        std::cout << " Mode:Normal" << std::endl;
    }
}

/**********************************************************************/
