/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "../../sim/network/incc.h"
#include "../../sim/network/router.h"
#include "../../sim/common/mainmemory.h"
#include "../../sim/common/config.h"
#include "../../sim/common/memorymap.h"
#include <iostream>

Incc incc[2];
MainMemory mem[2];
Router rt[4];
SystemConfig sc;

/**********************************************************************/
bool anybusy()
{
    if (incc[0].isbusy() || incc[1].isbusy())
        return true;
    for (int i = 0; i < 4; i++)
        if (rt[i].isbusy())
            return true;
    return false;
}

/**********************************************************************/
int dma_wait()
{
    int cnt = 0;
    while (anybusy() && cnt < 1000) {
        cnt++;
        incc[0].drive();
        incc[1].drive();
        for (int i = 0; i < 4; i++)
            rt[i].drive();
        incc[0].update();
        incc[1].update();
        for (int i = 0; i < 4; i++)
            rt[i].update();
/*
        // For DEBUG
        for (int i = 0; i < 4; i++) {
            if (rt[i].isbusy()) {
                std::cout << "cnt:" << cnt << " rt:" << i << std::endl;
                rt[i].print();
            }
        }
*/
    }
    return cnt;
}

/**********************************************************************/
void dma_put(int src, ullint laddr, ullint raddr,
             uint lstride, uint rstride, uint size)
{
    if (src == 0) { // PUT32
        mem[0].write(SC_MMAP_DMA_PUT_REMOTE_ID    , 4, 0x101);
        mem[0].write(SC_MMAP_DMA_PUT_REMOTE_ADDR  , 4, raddr);
        mem[0].write(SC_MMAP_DMA_PUT_REMOTE_STRIDE, 4, rstride);
        mem[0].write(SC_MMAP_DMA_PUT_LOCAL_ADDR   , 4, laddr);
        mem[0].write(SC_MMAP_DMA_PUT_LOCAL_STRIDE , 4, lstride);
        mem[0].write(SC_MMAP_DMA_PUT_SIZE         , 4, size);
        mem[0].write(SC_MMAP_DMA_PUT_CMD          , 4, DMA_PUT);
    } else {        // PUT64
        mem[1].write(SC_MMAP_DMA_PUT_REMOTE_ID    , 4, 0x000);
        mem[1].write(SC_MMAP_DMA_PUT_REMOTE_ADDR  , 4, raddr);
        mem[1].write(SC_MMAP_DMA_PUT_REMOTE_STRIDE, 4, raddr >> 32);
        mem[1].write(SC_MMAP_DMA_PUT_LOCAL_ADDR   , 4, laddr);
        mem[1].write(SC_MMAP_DMA_PUT_LOCAL_STRIDE , 4, lstride);
        mem[1].write(SC_MMAP_DMA_PUT_SIZE         , 4, size);
        mem[1].write(SC_MMAP_DMA_PUT_CMD          , 4, DMA_PUT64);
    }
}

/**********************************************************************/
void dma_get(int src, ullint laddr, ullint raddr,
             uint lstride, uint rstride, uint size)
{
    if (src == 0) { // GET32
        mem[0].write(0x800, 4, 0x000);
        mem[0].write(0x804, 4, raddr);
        mem[0].write(0x808, 4, rstride);
        mem[0].write(0x80c, 4, laddr);
        mem[0].write(0x810, 4, lstride);
        mem[0].write(0x814, 4, size);
        mem[0].write(0x818, 4, 0);
        mem[0].write(0x81c, 4, DMA_PUT);
    } else {        // GET64
        mem[1].write(0x800, 4, 0x101);
        mem[1].write(0x804, 4, laddr);
        mem[1].write(0x808, 4, lstride);
        mem[1].write(0x80c, 4, raddr);
        mem[1].write(0x810, 4, raddr >> 32);
        mem[1].write(0x814, 4, size);
        mem[1].write(0x818, 4, 0);
        mem[1].write(0x81c, 4, DMA_PUT64);
    }
    dma_put(src, 0x800, SC_MMAP_DMA_GET_REMOTE_ID, 4, 4, 32);
}

/**********************************************************************/
void test ()
{
    mem[0].init(0x200000000ull);
    mem[1].init(0x10000);
    incc[0].init(0x0101, sc, &mem[0]);
    incc[1].init(0x0202, sc, &mem[1]);
    rt[0].init(0x0101, sc);
    rt[1].init(0x0201, sc);
    rt[2].init(0x0102, sc);
    rt[3].init(0x0202, sc);

    rt[0].connect(DIR_E, rt[1], opposite(DIR_E));
    rt[0].connect(DIR_S, rt[2], opposite(DIR_S));
    rt[1].connect(DIR_W, rt[0], opposite(DIR_W));
    rt[1].connect(DIR_S, rt[3], opposite(DIR_S));
    rt[2].connect(DIR_E, rt[3], opposite(DIR_E));
    rt[2].connect(DIR_N, rt[0], opposite(DIR_N));
    rt[3].connect(DIR_W, rt[2], opposite(DIR_W));
    rt[3].connect(DIR_N, rt[1], opposite(DIR_N));
    incc[0].connect(0, rt[0], DIR_C);
    incc[1].connect(0, rt[3], DIR_C);
    rt[0].connect(DIR_C, incc[0], 0);
    rt[3].connect(DIR_C, incc[1], 0);
    for (uint addr = 0x1000; addr < 0x1040; addr += 4) {
        mem[0].write(addr, 4, addr);
        mem[1].write(addr, 4, addr);
    }
    mem[0].write(SC_MMAP_DMA_GET_CH, 4, 1);
    mem[1].write(SC_MMAP_DMA_GET_CH, 4, 1);

    dma_get(0, 0x2000, 0x1010, 4, 4, 40);
    int cycle = dma_wait();

    std::cout << "Total Cycle: " << cycle << std::endl;
    std::cout << "Memory Check: ";
    bool check = true;
    for (int i = 0; i < 12; i++) {
        int buf;
        mem[0].read(0x2000 + i * 4, 4, &buf);
        check = check && (buf == ((i < 10) ? 0x1010 + i * 4 : 0));
    }
    std::cout << ((check) ? "PASS" : "FAIL") << std::endl << std::endl;
}

/**********************************************************************/
int main (void)
{
    sc.setdefaultconfig(1, 1, true, false);
    sc.num_channel  = 2;
    for (int i = 0; i < 64; i++) {
        sc.num_inbuf    = (i & 1) ? 8 : 4;
        sc.lookahead_rc = (i & 2);
        sc.fix_vc       = (i & 4);
        sc.fix_priority = (i & 8);
        sc.credit_fc    = (i & 16);
        sc.skip_lt      = (i & 32);
        std::cout << "Buf: " << sc.num_inbuf;
        if (sc.lookahead_rc)
            std::cout << " NRC";
        if (sc.fix_vc)
            std::cout << " F_VC";
        if (sc.fix_priority)
            std::cout << " F_PR";
        if (sc.credit_fc)
            std::cout << " CRED";
        if (sc.skip_lt)
            std::cout << " NOLT";
        std::cout << std::endl;
        test();
    }

    return 0;
}

/**********************************************************************/










