/**********************************************************************/
/* SimMc: Simulator for Many-Core Architecture Research Environment   */
/* Copyright (C) 2010-2013 Arch Lab. Tokyo Institute of Technology    */
/*  Licensed under GPL. See COPYING file for more information.        */
/**********************************************************************/
#include "../../sim/network/incc.h"
#include "../../sim/network/router.h"
#include "../../sim/smart/smartunit.h"
#include "../../sim/common/mainmemory.h"
#include "../../sim/common/config.h"
#include "../../sim/common/memorymap.h"
#include <iostream>

Incc incc[4];
MainMemory mem[4];
SmartUnit smart[4];
Router rt[4];
SystemConfig sc;

/**********************************************************************/
bool anybusy()
{
    for (int i = 0; i < 4; i++)
        if (incc[i].isbusy() || rt[i].isbusy())
            return true;
    return false;
}

/**********************************************************************/
int dma_wait()
{
    int cnt = 0;
    while (anybusy() && cnt < 1000) {
        cnt++;
        for (int i = 0; i < 4; i++) {
            incc[i].drive();
            rt[i].drive();
            smart[i].drive();
        }
        for (int i = 0; i < 4; i++) {
            incc[i].update();
            rt[i].update();
            smart[i].update();
        }
/*
        // For DEBUG
        for (int i = 0; i < 4; i++) {
            if (rt[i].isbusy()) {
                std::cout << "cnt:" << cnt << " rt:" << i << std::endl;
                rt[i].print();
            }
        }
*/
    }
    return cnt;
}

/**********************************************************************/
void dma_put(int src, ullint laddr, ullint raddr,
             uint lstride, uint rstride, uint size)
{
    mem[src].write(SC_MMAP_DMA_PUT_REMOTE_ID    , 4, 0x201);
    mem[src].write(SC_MMAP_DMA_PUT_REMOTE_ADDR  , 4, raddr);
    mem[src].write(SC_MMAP_DMA_PUT_REMOTE_STRIDE, 4, rstride);
    mem[src].write(SC_MMAP_DMA_PUT_LOCAL_ADDR   , 4, laddr);
    mem[src].write(SC_MMAP_DMA_PUT_LOCAL_STRIDE , 4, lstride);
    mem[src].write(SC_MMAP_DMA_PUT_SIZE         , 4, size);
    mem[src].write(SC_MMAP_DMA_PUT_CMD          , 4, DMA_PUT);
}

/**********************************************************************/
void dma_get(int src, ullint laddr, ullint raddr,
             uint lstride, uint rstride, uint size)
{
    mem[src].write(0x800, 4, 0x101);
    mem[src].write(0x804, 4, raddr);
    mem[src].write(0x808, 4, rstride);
    mem[src].write(0x80c, 4, laddr);
    mem[src].write(0x810, 4, lstride);
    mem[src].write(0x814, 4, size);
    mem[src].write(0x818, 4, 0);
    mem[src].write(0x81c, 4, DMA_PUT);
    dma_put(src, 0x800, SC_MMAP_DMA_GET_REMOTE_ID, 4, 4, 32);
}

/**********************************************************************/
void test ()
{
    uint index2id[4] = {0x0101, 0x0201, 0x0102, 0x0202};
    for (int i = 0; i < 4; i++) {
        mem[i].init(0x10000);
        incc[i].init (index2id[i], sc, &mem[i]);
        smart[i].init(index2id[i], sc, &rt[i]);
        rt[i].init   (index2id[i], sc);
    }

    rt[0].connect(DIR_E, rt[1], opposite(DIR_E));
    rt[0].connect(DIR_S, rt[2], opposite(DIR_S));
    rt[1].connect(DIR_W, rt[0], opposite(DIR_W));
    rt[1].connect(DIR_S, rt[3], opposite(DIR_S));
    rt[2].connect(DIR_E, rt[3], opposite(DIR_E));
    rt[2].connect(DIR_N, rt[0], opposite(DIR_N));
    rt[3].connect(DIR_W, rt[2], opposite(DIR_W));
    rt[3].connect(DIR_N, rt[1], opposite(DIR_N));
    for (int i = 0; i < 4; i++) {
        incc[i].connect(0, smart[i], SmartUnit::SIDE_INCC);
        smart[i].connect(SmartUnit::SIDE_INCC, incc[i], 0);
        smart[i].connect(SmartUnit::SIDE_RT, rt[i], DIR_C);
        rt[i].connect(DIR_C, smart[i], SmartUnit::SIDE_RT);
        for (uint addr = 0x1000; addr < 0x1040; addr += 4) {
            mem[i].write(addr, 4, addr);
        }
    }

    dma_get(0, 0x2000, 0x1010, 4, 4, 40);
    if (sc.ismaster(0x0101))
        dma_get(1, 0x2000, 0x1010, 4, 4, 40);
    int cycle = dma_wait();

    std::cout << "Total Cycle: " << cycle << std::endl;
    std::cout << "Memory Check: ";
    bool check = true;
    for (int i = 0; i < 12; i++) {
        int buf;
        mem[0].read(0x2000 + i * 4, 4, &buf);
        check = check && (buf == ((i < 10) ? 0x1010 + i * 4 : 0));
    }
    std::cout << ((check) ? "PASS" : "FAIL") << std::endl << std::endl;
}

/**********************************************************************/
int main (void)
{
    const char *mapfile[4] = {"map1.txt", "map2.txt",
                              "map3.txt", "map4.txt"};
    sc.num_channel  = 3;
    sc.fix_vc       = true;
    sc.fix_priority = false;
    for (int i = 0; i < 64; i++) {
        sc.readconfigfile(mapfile[i % 4]);
        sc.num_inbuf    = (i & 4) ? 8 : 4;
        sc.lookahead_rc = (i & 8);
        sc.credit_fc    = (i & 16);
        sc.skip_lt      = (i & 32);
        std::cout << mapfile[i % 4] << " ";
        std::cout << "Buf: " << sc.num_inbuf;
        if (sc.lookahead_rc)
            std::cout << " NRC";
        if (sc.credit_fc)
            std::cout << " CRED";
        if (sc.skip_lt)
            std::cout << " NOLT";
        std::cout << std::endl;
        test();
    }

    return 0;
}

/**********************************************************************/










